/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.clusterContext.contextAnalysis;

import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.*;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.PolicyAnalyser;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.QoSPolicyViolationEvaluator;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.ServerLoadPolicyViolationEvaluator;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.ContextEntropyEvaluator;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.RLNode;
import ro.utcn.dsrl.utils.logging.GlobalLoopLogger;
import ro.utcn.dsrl.utils.misc.Pair;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.atomic.AtomicLong;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */


public class ClusterContextEntropyEvaluation implements ContextEntropyEvaluator {

    private PolicyAnalyser policyAnalyser;
    private ServerLoadPolicyViolationEvaluator serverLoadViolationEvaluator;
    private QoSPolicyViolationEvaluator qosPolicyViolationEvaluator;


    public ClusterContextEntropyEvaluation(PolicyAnalyser policyAnalyser,
                                           ServerLoadPolicyViolationEvaluator serverLoadViolationEvaluator,
                                           QoSPolicyViolationEvaluator qosPolicyViolationEvaluator) {
        super();
        this.policyAnalyser = policyAnalyser;
        this.serverLoadViolationEvaluator = serverLoadViolationEvaluator;
        this.qosPolicyViolationEvaluator = qosPolicyViolationEvaluator;
    }


    public Pair<Double, List<GPI_KPI_Policy>> computeEntropy(Collection<GPI_KPI_Policy> policies) {

        final List<GPI_KPI_Policy> brokenPolicies = new Vector<GPI_KPI_Policy>();

        final AtomicLong entropy = new AtomicLong(0);

        Collection<ITComputingContextPolicy> itComputingContextPolicies = new ArrayList<ITComputingContextPolicy>();
        Collection<QoSPolicy> qoSPolicies = new ArrayList<QoSPolicy>();
        for (GPI_KPI_Policy gpi_kpi_policy : policies) {
            if (gpi_kpi_policy instanceof QoSPolicy) {
                qoSPolicies.add((QoSPolicy) gpi_kpi_policy);
            } else if (gpi_kpi_policy instanceof ITComputingContextPolicy) {
                itComputingContextPolicies.add((ITComputingContextPolicy) gpi_kpi_policy);
            } else {
                GlobalLoopLogger.getLogger().info("This should not happen. Unknown class for policy " + gpi_kpi_policy);
                Exception e = new Exception("This should not happen. Unknown class for policy " + gpi_kpi_policy);
                e.fillInStackTrace();
                e.printStackTrace();
            }
        }

        for (QoSPolicy policy : qoSPolicies) {

            Collection<ContextResource> task = policy.getPolicySubject();

            // if task has been deleted
            if (task == null) {
                continue;
            }

            if (!policyAnalyser.isPolicyRespected(policy)) {

                brokenPolicies.add(policy);

                if (policy.hasPolicyWeight()) {
                    for (ContextResource app : task) {
                        entropy.addAndGet(new Double(policy.getPolicyWeight()
                                * qosPolicyViolationEvaluator.evaluateQoSPolicyViolationDegree((ApplicationActivity) app)).longValue());
                    }
                }
            }
        }


        for (ITComputingContextPolicy policy : itComputingContextPolicies) {
            ServiceCenterServer serviceCenterServer = (ServiceCenterServer) policy.getPolicySubject().iterator().next();


            if (!serviceCenterServer.isCluster()) {

                Collection<ContextResource> servers = policy.getPolicySubject();
                for (ContextResource r : servers) {

                    if (serviceCenterServer.getIsActive()) {

                        if (!policyAnalyser.isPolicyRespected(policy)) {

                            brokenPolicies.add(policy);

                            if (policy.hasPolicyWeight()) {
                                entropy.addAndGet(new Double(policy.getPolicyWeight()
                                        * serverLoadViolationEvaluator.evaluateServerLoadPolicyViolationDegree(serviceCenterServer)).longValue());
                            } else {
                                entropy.addAndGet(new Double(serverLoadViolationEvaluator.evaluateServerLoadPolicyViolationDegree(serviceCenterServer)).longValue());
                            }

                        }

                    }
                }
            } else {
                Collection<ServiceCenterITComputingResource> contextResources = serviceCenterServer.getResources();

                Collection<ServiceCenterServer> attachedComplexResources = new ArrayList<ServiceCenterServer>(contextResources.size());
                for (ContextResource resource : contextResources) {
                    if (resource instanceof ServiceCenterServer) {
                        attachedComplexResources.add((ServiceCenterServer) resource);
                    }
                }

                //barrier for synchronizing the compute entropy threads

//                final CountDownLatch countDownLatch = new CountDownLatch(attachedComplexResources.size());
                //compute entropy  contribution for each service center server (or cluster)
                for (final ServiceCenterServer resource : attachedComplexResources) {
//
//                    Thread thread = new Thread() {
//                        @Override
//                        public void run() {
                    ArrayList<GPI_KPI_Policy> policiesForEntropyComputation = new ArrayList<GPI_KPI_Policy>();
                    ITComputingContextPolicy serviceCenterPolicy = resource.getServiceCenterServerPolicy();
                    policiesForEntropyComputation.add(serviceCenterPolicy);

                    //se calculeaza pentru frunze\(servere nu clustere)
//                    for (ApplicationActivity applicationActivity : resource.getRunningActivities()) {
//                        GPI_KPI_Policy gpi_kpi_policy = (GPI_KPI_Policy) applicationActivity.getActivityPolicies().iterator().next();
//                        policiesForEntropyComputation.add(gpi_kpi_policy);
//                    }
                    if (!policyAnalyser.isPolicyRespected(serviceCenterPolicy)) {
                        Pair<Double, List<GPI_KPI_Policy>> result = computeEntropy(policiesForEntropyComputation);
                        brokenPolicies.add(serviceCenterPolicy);
                        entropy.addAndGet(new Double(result.getFirst()).longValue());
                    }
//                            countDownLatch.countDown();
//                        }
//                    };
//                    thread.start();
                }

//                try {
//                    countDownLatch.await();
//                } catch (InterruptedException e) {
//                    ro.utcn.dsrl.utils.logging.GlobalLoopLogger.error(e);
//                }
            }
        }
        return new Pair<Double, List<GPI_KPI_Policy>>(entropy.doubleValue(), brokenPolicies);
    }

    public double computeRewardFunction(RLNode previous,
                                        RLNode current, List<ContextAction> contextActions) {
        double function;

        if (previous != null && previous.getActions().size() > 0) {
            function = previous.getRewardFunction();
        } else {
            function = 0;
        }
        if (previous != null) {
            int cost = 0;
            for (ContextAction action : contextActions) {
                cost += action.getCost();
            }
            double temp = previous.getContextEntropy()
                    - current.getContextEntropy() - cost; // - c.getCost() -
            // current.getActions().size()
            // * 100;
            function += RLNode.gamma * temp;

        } else {
            function = -current.getContextEntropy();
        }

        return function;
    }
}
