/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.clusterContext.contextAnalysis;

import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.*;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.ServerLoadPolicyViolationEvaluator;
import ro.utcn.dsrl.utils.config.GeneralConfigurationManager;

import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicLong;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */


public class ClusterServerLoadPolicyViolationEvaluator implements ServerLoadPolicyViolationEvaluator {

    public double evaluateServerLoadPolicyViolationDegree(ServiceCenterServer server) {
        Collection<ServiceCenterITComputingResource> serviceCenterITComputingResources = server.getResources();

        if (!server.isCluster()) {

            double respectance = 0.0;
            Collection<ServiceCenterITComputingResource> resources = server
                    .getResources();

            CPU associatedCPU = null;
            HDD storage = null;
            MEM memory = null;
            double diff = 0;
            if (server.getIsActive()) {
                for (CPU cpu : server.getCpuResources()) {
                    Double usedCPU = cpu.getCurrentWorkLoad();
                    Double totalCPU = cpu.getMaximumWorkLoad();
                    Double optimalCPU = cpu.getOptimalWorkLoad();
                    Double cpuMaxAcceptableValue =  optimalCPU + GeneralConfigurationManager.getResourceLoadVariation()*totalCPU/100;
;
                    Double cpuMinAcceptableValue = optimalCPU - GeneralConfigurationManager.getResourceLoadVariation()*totalCPU/100;
                    if (usedCPU < cpuMinAcceptableValue) {
                        diff += cpuMinAcceptableValue - usedCPU;
                    }
                    if (usedCPU > cpuMaxAcceptableValue) {
                        diff += usedCPU - cpuMaxAcceptableValue;
                    }
                    if (server.hasCPUWeight()) {
                        respectance += server.getCPUWeight() * diff;
                    } else {
                        respectance += diff;
                    }
                }

                diff = 0.0;
                double usedMemory = 0.0;
                double total = 0.0;
                double optimal = 0.0;
                double memoryMaxAcceptableValue = 0.0;

                double memoryMinAcceptableValue = optimal / 2.0;
                Collection<MEM> memories = server.getMemResources();
                for (MEM mem : memories) {
                    usedMemory = mem.getCurrentWorkLoad();
                    total = mem.getMaximumWorkLoad();
                    optimal = mem.getOptimalWorkLoad();
                    memoryMaxAcceptableValue = optimal + GeneralConfigurationManager.getResourceLoadVariation()*total/100;
                    memoryMinAcceptableValue = optimal - GeneralConfigurationManager.getResourceLoadVariation()*total/100;
                    if (usedMemory > memoryMaxAcceptableValue) {
                        diff = usedMemory - memoryMaxAcceptableValue;
                    } else if (usedMemory < memoryMinAcceptableValue) {
                        diff = memoryMinAcceptableValue - usedMemory;
                    }
                    if (server.hasMEMWeight()) {
                        respectance += server.getMEMWeight() * diff;
                    } else {
                        respectance += diff;
                    }
                }
                diff = 0.0;

                Double usedStorage = 0.0;
                Double storageMaxAcceptableValue = 0.0;
                double storageMinAcceptableValue = 0.0;

                for (HDD hdd : server.getHddResources()) {
                    total = hdd.getMaximumWorkLoad();
                    optimal = hdd.getOptimalWorkLoad();
                    usedStorage = hdd.getCurrentWorkLoad();
                    storageMaxAcceptableValue = (2 * total + optimal) / 3.0;
                    storageMinAcceptableValue = optimal / 2.0;
                    if (usedStorage > storageMaxAcceptableValue) {
                        diff = usedStorage - storageMaxAcceptableValue;
                    } else if (usedStorage < storageMinAcceptableValue) {
                        diff = storageMinAcceptableValue - usedStorage;
                    }
                    if (server.hasHDDWeight()) {
                        respectance += server.getHDDWeight() * diff;
                    } else {
                        respectance += diff;
                    }
                }
            }
            return respectance;
        } else {
            Collection<ServiceCenterServer> attachedComplexResources = new ArrayList<ServiceCenterServer>(serviceCenterITComputingResources.size());
            for (ServiceCenterITComputingResource resource : serviceCenterITComputingResources) {
                if (resource instanceof ServiceCenterServer) {
                    attachedComplexResources.add((ServiceCenterServer) resource);
                }
            }
            //make respectance value thread safe
            final AtomicLong respectance = new AtomicLong();

            //barrier for synchronizing the compute entropy threads
            final CountDownLatch countDownLatch = new CountDownLatch(attachedComplexResources.size());

            //compute entropy  contribution for each running activity
            for (final ServiceCenterServer resource : attachedComplexResources) {
                Thread thread = new Thread("Entropy contribution thread for " + resource) {
                    @Override
                    public void run() {
                        respectance.addAndGet(new Double(evaluateServerLoadPolicyViolationDegree(resource)).longValue());
                        countDownLatch.countDown();
                    }
                };
                thread.start();
            }

            try {
                countDownLatch.await();
            } catch (InterruptedException e) {
                ro.utcn.dsrl.utils.logging.GlobalLoopLogger.error(e);
            }
            return 0;
        }
    }

     public double getMinWorkload(SimpleResource simpleResource) {
          Double total = simpleResource.getMaximumWorkLoad();
          Double optimal = simpleResource.getOptimalWorkLoad();
          Double minAcceptableValue =  optimal - GeneralConfigurationManager.getResourceLoadVariation()*total/100;

          return minAcceptableValue;
    }

    public double getMaxWorkload(SimpleResource simpleResource) {
                Double total = simpleResource.getMaximumWorkLoad();
                Double optimal = simpleResource.getOptimalWorkLoad();
                Double maxAcceptableValue = optimal + GeneralConfigurationManager.getResourceLoadVariation()*total/100;
          return maxAcceptableValue;
    }
}
