/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.actionEnforcementModule.agents.behaviors;

import edu.stanford.smi.protege.model.KnowledgeBase;
import jade.core.behaviours.OneShotBehaviour;
import org.apache.log4j.Logger;
import ro.utcn.dsrl.actionEnforcementModule.actionEnforcement.ContextActionEnforcer;
import ro.utcn.dsrl.actionEnforcementModule.agents.ActionPlanEnforcementAgent;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.CAPInteractions;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.util.BPELCallUtil;
import ro.utcn.dsrl.actionPlanningModule.utils.actionExecutionSimulation.IActionSimulationUtil;
import ro.utcn.dsrl.actionPlanningModule.utils.actionExecutionSimulation.impl.DefaultActionSimulationUtil;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextAccessAndProcessing.ContextAccessAndProcessingAPI;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.api.EACMInstance;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.*;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.ContextAction;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.util.exceptions.ContextModelAccessException;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.impl.PowerConsumptionEstimator;
import ro.utcn.dsrl.contextMonitoringModule.agents.ContextMonitoringAgent;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.ActionToDtoMapper;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.ServerToDtoMapper;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.TaskToDtoMapper;
import ro.utcn.dsrl.utils.actionEnforcement.ActionPlanDto;
import ro.utcn.dsrl.utils.agentsCommunication.AgentMessageType;
import ro.utcn.dsrl.utils.agentsCommunication.InterAgentMessageBuilder;
import ro.utcn.dsrl.utils.commonDtos.*;
import ro.utcn.dsrl.utils.commonInterfaces.Interruptable;
import ro.utcn.dsrl.utils.commonInterfaces.PowerConsumption;
import ro.utcn.dsrl.utils.config.GeneralConfigurationManager;
import ro.utcn.dsrl.utils.exceptions.ApplicationException;
import ro.utcn.dsrl.utils.logging.GlobalLoopLogger;
import ro.utcn.dsrl.workloadScheduler.WorkloadSchedulerAgent;

import java.rmi.RemoteException;
import java.util.*;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */
public class APEAExecuteActionPlanBehavior extends OneShotBehaviour implements
        Interruptable {

    private EACMInstance eacmInstance;
    private ContextAccessAndProcessingAPI contextAccessAndProcessingAPI;
    private ContextActionEnforcer actionEnforcer;
    private CAPInteractions capInteractions;
    private Logger logger;
    private ActionPlanDto actionPlanDto;
    private Thread currentlyRunningThread;
    private ActionPlanEnforcementAgent actionPlanEnforcementAgent;
    private IActionSimulationUtil simulationUtil;

    public APEAExecuteActionPlanBehavior(ActionPlanEnforcementAgent actionPlanEnforcementAgent, EACMInstance eacmInstance,
            ContextAccessAndProcessingAPI contextAccessAndProcessingAPI,
            ContextActionEnforcer actionEnforcer, CAPInteractions capInteractions,
            ActionPlanDto actionPlanDto, Logger logger) {
        super(actionPlanEnforcementAgent);
        this.actionPlanEnforcementAgent = actionPlanEnforcementAgent;
        this.eacmInstance = eacmInstance;
        this.contextAccessAndProcessingAPI = contextAccessAndProcessingAPI;
        this.actionEnforcer = actionEnforcer;
        this.capInteractions = capInteractions;
        this.logger = logger;
        this.actionPlanDto = actionPlanDto;
        simulationUtil = new DefaultActionSimulationUtil(null);
    }

    public void action() {
        currentlyRunningThread = new Thread("Executing action plan thread") {

            public void run() {
                logger.info("Executing action plan");
                List<String> actionPlanActions = actionPlanDto.getActions();
                Queue<ContextAction> actionsToBeExecuted = new LinkedList<ContextAction>();
                for (String actionName : actionPlanActions) {
                    ContextAction action = eacmInstance.getITComputingResourceAdaptationAction(actionName);
                    actionsToBeExecuted.add(action);
                    GlobalLoopLogger.getLogger().info("Executing " + action.toString());
                }

                if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                    if (actionsToBeExecuted.size() > 0) {

                        ContextSituationDto situationDto = new ContextSituationDto();
                        PowerConsumption powerConsumption = null;
                        try {
                            powerConsumption = new PowerConsumptionEstimator(eacmInstance);
                        } catch (RemoteException e) {
                            GlobalLoopLogger.error(e);
                        }

                        try {
                            situationDto.setEnergyLevel(powerConsumption.getValueWithRunningAlgorithm().doubleValue());
                        } catch (Exception e) {
                            situationDto.setEnergyLevel(1000000000);
                            GlobalLoopLogger.getLogger().debug(e.toString());
                        }
                        Collection<ServiceCenterServer> serversList = eacmInstance.getAllServiceCenterServerInstances();
                        Collection<ApplicationActivity> actionPlanTasks = eacmInstance.getAllApplicationActivityInstances();

                        for (ServiceCenterServer server : serversList) {
                            situationDto.addServerDto(ServerToDtoMapper.map(server));
                        }

                        for (ApplicationActivity activity : actionPlanTasks) {
                            situationDto.addTaskDto(TaskToDtoMapper.map(activity));
                        }
                        for (ContextAction action : actionsToBeExecuted) {
                            situationDto.addActionDto(ActionToDtoMapper.map(action));
                        }

                        Collection<ContextSituationDto> contextSituationDtos = null;
                        try {
                            contextSituationDtos = contextAccessAndProcessingAPI.getHistoricalContextSituations();
                            contextSituationDtos.add(situationDto);
                            contextAccessAndProcessingAPI.persistHistoricalContextSituations(contextSituationDtos);
                        } catch (ContextModelAccessException e) {
                            GlobalLoopLogger.error(e);
                        }

                    }
                }
//                executeActionsLinearly(actionsToBeExecuted);
                
                List<List<Thread>> actionThreads = parallelizeActionsExecution(actionsToBeExecuted);
                for (List<Thread> parralelThreads : actionThreads) {
                    if (currentlyRunningThread.isInterrupted()) {
                        return;
                    }
                    for (Thread actionThread : parralelThreads) {
                        actionThread.start();
                    }
                    for (Thread actionThread : parralelThreads) {
                        if (currentlyRunningThread.isInterrupted()) {
                            return;
                        }
                        if (actionThread.isAlive()) {
                            try {
                                actionThread.join();
                            } catch (InterruptedException ex) {
                                logger.info(ex.getMessage());
                                logger.debug(ex.toString());
                            }
                        }
                    }
                }

//            InterAgentMessageBuilder.sendMessage(actionPlanEnforcementAgent, WorkloadSchedulerAgent.class, AgentMessageType.ACTION_PLAN_EXECUTION_COMPLETED);

                if (GeneralConfigurationManager.isWSCallbackEnabled()) {
                    try {
                        BPELCallUtil.signalDepoloymentFailure("success");
                    } catch (Exception e) {
                        GlobalLoopLogger.getLogger().info(e.toString());
                    }
                }
                
                //delete executed actions to free memory
                for (ContextAction contextAction : eacmInstance.getAllConsolidationActionInstances()) {
                    contextAction.delete();
                }
                
                
            }
        };
        currentlyRunningThread.start();
        try {
            currentlyRunningThread.join();
        } catch (InterruptedException e) {
            logger.debug(e.toString());
            logger.info(e.getMessage());
        }


//        List<ConsolidationAction> consolidationActions = new ArrayList<ConsolidationAction>();
//        consolidationActions.addAll(eacmInstance.getAllDeployActivityInstances());
//        consolidationActions.addAll(eacmInstance.getAllMigrateActivityInstances());
//        consolidationActions.addAll(eacmInstance.getAllSetServerStateActivityInstances());
//

        //try to free up memory
        Runtime.getRuntime().gc();
        //System.err.println("MEMORYYY "  + (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory())/1024);

        InterAgentMessageBuilder.sendMessage(actionPlanEnforcementAgent, ContextMonitoringAgent.class, AgentMessageType.ACTION_PLAN_EXECUTION_COMPLETED);
        if (GeneralConfigurationManager.usesWorkloadGenerator()) {
            InterAgentMessageBuilder.sendMessage(actionPlanEnforcementAgent, WorkloadSchedulerAgent.class, AgentMessageType.ACTION_PLAN_EXECUTION_COMPLETED);
        }

    }

    private List<List<Thread>> parallelizeActionsExecution(
            Queue<ContextAction> actions) {
        List<List<Thread>> result = new ArrayList<List<Thread>>();
        List<ContextAction> actionsList = new ArrayList(actions);
        ListIterator<ContextAction> contextActionIterator = actionsList.listIterator();

        while (contextActionIterator.hasNext()) {
            result.add(extractNextParallelizableSequence(contextActionIterator));
        }
        return result;
    }

    private List<Thread> extractNextParallelizableSequence(
            ListIterator<ContextAction> contextActionIterator) {

        final ContextActionEnforcer actionEnforcer = this.actionEnforcer;

        if (!contextActionIterator.hasNext()) {
            return new ArrayList<Thread>();
        }
        ContextAction action = contextActionIterator.next();

//        if (!action.hasExecutedSubActions()) {
//            System.err.println("NO SUB-ACTIONS FOR " + action);
//        }

        if (action instanceof DeployActivity) {
            List<DeployActivity> deployActionsSubList = new ArrayList<DeployActivity>();
            deployActionsSubList.add((DeployActivity) action);
            while (contextActionIterator.hasNext()) {
                action = contextActionIterator.next();
                if (action instanceof DeployActivity) {
                    deployActionsSubList.add((DeployActivity) action);
                } else {
                    contextActionIterator.previous();
                    break;
                }

            }
            List<Thread> deploymentThreads = new ArrayList<Thread>(
                    deployActionsSubList.size());


            for (final DeployActivity deployAction : deployActionsSubList) {
                Thread thread = new Thread("Thread for " + deployAction.toString()) {

                    public void run() {

                        try {
                            simulationUtil.simulateActionExecution(eacmInstance, deployAction);

                            assert deployAction.getResourceTo().getRunningActivities().contains(deployAction.getActivity()) : "Invalid deploy action " + deployAction;

                            ApplicationActivity task = deployAction.getActivity();
                            ActivityDto activityDto = new ActivityDto();
                            activityDto.setEpkbID(task.getEPKBId());
                            activityDto.setIp(task.getNetworkInfo().getIpAddress());

                            String virtualMachineConf = (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) ? capInteractions.getService(activityDto).getVmConfiguration() : "";

                            if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                                capInteractions.addDeployedService(activityDto);
                            }
                            actionEnforcer.executeDeployActivityAction(deployAction, virtualMachineConf);


                        } catch (ApplicationException ex) {
                            logger.info("Error \" " + ex.getMessage()
                                    + "\" executing " + deployAction.toString());
                            logger.error(ex.toString());
                            ex.printStackTrace();
                        }

                    }
                };

                deploymentThreads.add(thread);
            }
            return deploymentThreads;
        } else if (action instanceof SetServerStateActivity) {
            List<ContextAction> setServerStateActions = new ArrayList<ContextAction>();
            setServerStateActions.add(action);
            while (contextActionIterator.hasNext()) {
                action = contextActionIterator.next();
                if (action instanceof SetServerStateActivity) {
                    setServerStateActions.add(action);
                } else {
                    contextActionIterator.previous();
                    break;
                }

            }

            List<Thread> setServerStateThreads = new ArrayList<Thread>(
                    setServerStateActions.size());
            for (final ContextAction setServerState : setServerStateActions) {
                Thread thread = new Thread("Thread for " + setServerState.toString()) {

                    public void run() {
                        try {
                            actionEnforcer.executeSetServerStateAction(setServerState);
                            simulationUtil.simulateActionExecution(eacmInstance, setServerState);
                        } catch (ApplicationException ex) {
                            logger.info("Error \" " + ex.getMessage()
                                    + "\" executing "
                                    + setServerState.toString());
                            logger.debug(ex.toString());
                        }
                    }
                };

                setServerStateThreads.add(thread);
            }
            return setServerStateThreads;
        } else if (action instanceof MigrateActivity) {
            final MigrateActivity migrateAction = (MigrateActivity) action;
            List<Thread> threads = new ArrayList<Thread>(1);
            Thread thread = new Thread("Thread for " + migrateAction.toString()) {

                public void run() {

                    try {

                        actionEnforcer.executeMigrateActivityAction(migrateAction);
                        simulationUtil.simulateActionExecution(eacmInstance, migrateAction);
                        assert migrateAction.getResourceTo().getRunningActivities().contains(migrateAction.getActivity()) : "Invalid deploy for migrate action " + migrateAction;
                        assert !migrateAction.getResourceFrom().getRunningActivities().contains(migrateAction.getActivity()) : "Invalid remove for migrate action " + migrateAction;
                    } catch (ApplicationException ex) {
                        logger.info("Error \" " + ex.getMessage()
                                + "\" executing " + migrateAction.toString());
                        logger.debug(ex.toString());
                    }

                }
            };
            threads.add(thread);
            return threads;
        }
        if (action instanceof RemoveActivity) {
            final RemoveActivity removeActivity = (RemoveActivity) action;
            List<Thread> threads = new ArrayList<Thread>(1);
            Thread thread = new Thread("Thread for " + removeActivity.toString()) {

                public void run() {
                    try {

                        simulationUtil.simulateActionExecution(eacmInstance, removeActivity);

                        ApplicationActivity task = removeActivity.getActivity();
                        ActivityDto activityDto = new ActivityDto();
                        activityDto.setEpkbID(task.getEPKBId());
                        activityDto.setIp(task.getNetworkInfo().getIpAddress());
                        actionEnforcer.executeRemoveActivityAction(removeActivity);

                        capInteractions.removeDeployedService(activityDto);
                    } catch (ApplicationException ex) {
                        logger.info("Error \" " + ex.getMessage()
                                + "\" executing " + removeActivity.toString());
                        logger.debug(ex.toString());
                    }
                }
            };
            threads.add(thread);
            return threads;
        } else {
            logger.info("Unrecognized action " + action.toString());
            return new ArrayList<Thread>();
        }

    }

    public void terminate() {
        if (currentlyRunningThread != null && currentlyRunningThread.isAlive()) {
            currentlyRunningThread.interrupt();
            try {
                currentlyRunningThread.join();
            } catch (InterruptedException e) {
                GlobalLoopLogger.error(e);
            }
        }
    }

    public void executeActionsLinearly(Queue<ContextAction> actions) {
        for (ContextAction action : actions) {

            if (action instanceof DeployActivity) {

                DeployActivity deployAction = (DeployActivity) action;
                try {
                    simulationUtil.simulateActionExecution(eacmInstance, deployAction);

                    assert deployAction.getResourceTo().getRunningActivities().contains(deployAction.getActivity()) : "Invalid deploy action " + deployAction;

                    ApplicationActivity task = deployAction.getActivity();
                    ActivityDto activityDto = new ActivityDto();
                    activityDto.setEpkbID(task.getEPKBId());
                    activityDto.setIp(task.getNetworkInfo().getIpAddress());

                    String virtualMachineConf = (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) ? capInteractions.getService(activityDto).getVmConfiguration() : "";
                    VirtualTaskInfo virtualTaskInfo = actionEnforcer.executeDeployActivityAction(deployAction, virtualMachineConf);
                    if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                        activityDto.setCloudID(virtualTaskInfo.getId());
                        capInteractions.addDeployedService(activityDto);
                    }

                } catch (ApplicationException ex) {
                    logger.info("Error \" " + ex.getMessage()
                            + "\" executing " + deployAction.toString());
                    logger.error(ex.toString());
                    ex.printStackTrace();
                }

            }
            if (action instanceof SetServerStateActivity) {
                SetServerStateActivity setServerState = (SetServerStateActivity) action;
                try {
                    actionEnforcer.executeSetServerStateAction(setServerState);
                    simulationUtil.simulateActionExecution(eacmInstance, setServerState);
                } catch (ApplicationException ex) {
                    logger.info("Error \" " + ex.getMessage()
                            + "\" executing "
                            + setServerState.toString());
                    logger.debug(ex.toString());
                }
            }
            if (action instanceof MigrateActivity) {
                MigrateActivity migrateAction = (MigrateActivity) action;
                try {
                    ApplicationActivity task = migrateAction.getActivity();
                    ActivityDto activityDto = new ActivityDto();
                    activityDto.setEpkbID(task.getEPKBId());
                    activityDto.setIp(task.getNetworkInfo().getIpAddress());
                    actionEnforcer.executeMigrateActivityAction(migrateAction);
                    simulationUtil.simulateActionExecution(eacmInstance, migrateAction);
                    assert migrateAction.getResourceTo().getRunningActivities().contains(migrateAction.getActivity()) : "Invalid deploy for migrate action " + migrateAction;
                    assert !migrateAction.getResourceFrom().getRunningActivities().contains(migrateAction.getActivity()) : "Invalid remove for migrate action " + migrateAction;
                } catch (ApplicationException ex) {
                    logger.info("Error \" " + ex.getMessage()
                            + "\" executing " + migrateAction.toString());
                    logger.debug(ex.toString());
                }
            }

            if (action instanceof RemoveActivity) {
                RemoveActivity removeActivity = (RemoveActivity) action;
                try {
                    actionEnforcer.executeMigrateActivityAction(removeActivity);
                    simulationUtil.simulateActionExecution(eacmInstance, removeActivity);

                    ApplicationActivity task = removeActivity.getActivity();
                    ActivityDto activityDto = new ActivityDto();
                    activityDto.setEpkbID(task.getEPKBId());
                    activityDto.setIp(task.getNetworkInfo().getIpAddress());

                    capInteractions.removeDeployedService(activityDto);
                } catch (ApplicationException ex) {
                    logger.info("Error \" " + ex.getMessage()
                            + "\" executing " + removeActivity.toString());
                    logger.debug(ex.toString());
                }
            }
        }
    }
}
