/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.actionPlanningModule.agents.behaviors;

import jade.core.Agent;
import jade.core.behaviours.OneShotBehaviour;
import java.io.IOException;
import ro.utcn.dsrl.actionEnforcementModule.agents.ActionPlanEnforcementAgent;
import ro.utcn.dsrl.actionPlanningModule.reasoning.PlanningEngine;
import ro.utcn.dsrl.actionPlanningModule.reasoning.impl.FFDPlanningEngine;
import ro.utcn.dsrl.actionPlanningModule.utils.actionExecutionSimulation.IActionSimulationUtil;
import ro.utcn.dsrl.actionPlanningModule.utils.actionExecutionSimulation.impl.DefaultActionSimulationUtil;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.ContextSituationToDtoMapper;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.RLNode;
import ro.utcn.dsrl.actionPlanningModule.reasoning.impl.ReinforcementLearningPlanningEngine;
import ro.utcn.dsrl.utils.similarContextIdentification.SimilarContextSituationIdentification;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.api.EACMInstance;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.*;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.impl.DefaultDeployActivity;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.impl.DefaultMigrateActivity;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.impl.DefaultSetServerStateActivity;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.utils.knowledgeBaseInterraction.impl.ContextSituationAccess;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.ContextEntropyEvaluator;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.impl.PowerConsumptionEstimator;
import ro.utcn.dsrl.utils.actionEnforcement.ActionPlanDto;
import ro.utcn.dsrl.utils.agentsCommunication.AgentMessageType;
import ro.utcn.dsrl.utils.agentsCommunication.InterAgentMessageBuilder;
import ro.utcn.dsrl.utils.commonDtos.*;
import ro.utcn.dsrl.utils.commonInterfaces.PowerConsumption;
import ro.utcn.dsrl.utils.commonInterfaces.Interruptable;
import ro.utcn.dsrl.utils.config.GeneralConfigurationManager;
import ro.utcn.dsrl.utils.exceptions.ApplicationException;
import ro.utcn.dsrl.utils.logging.GlobalLoopLogger;
import ro.utcn.dsrl.utils.misc.Pair;
import ro.utcn.dsrl.utils.misc.SimpleCSVWriter;

import java.rmi.RemoteException;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.*;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.CAPInteractions;
import ro.utcn.dsrl.actionPlanningModule.utils.ActionPlan;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.ActionToDtoMapper;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.ServerToDtoMapper;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.TaskToDtoMapper;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */
public class APAPlanningBehavior extends OneShotBehaviour implements Interruptable {

    private Agent agent;
    private ContextEntropyEvaluator entropyEvaluator;
    private PlanningEngine planningEngine;
    private EACMInstance eacmInstance;
    private Thread currentlyRunningThread;
    private SimpleCSVWriter simpleCSVWriter;
    private static final DateFormat dateFormat = new SimpleDateFormat("_dd_MM_HH_mm_");
    private SimilarContextSituationIdentification similarContextSituationIdentification;
    private CAPInteractions capInteractions;

    public APAPlanningBehavior(Agent agent, EACMInstance eacmInstance,
            ContextEntropyEvaluator entropyEvaluator, PlanningEngine planningEngine,
            CAPInteractions capInteractions) {
        super(agent);
        this.capInteractions = capInteractions;
        this.agent = agent;
        this.entropyEvaluator = entropyEvaluator;
        this.planningEngine = planningEngine;
        this.eacmInstance = eacmInstance;
        simpleCSVWriter = SimpleCSVWriter.getInstance(this.getClass());
//        loadMemory();

    }

    public void action() {
        currentlyRunningThread = new Thread("PAPlanningBehaviorThread") {

            public void run() {
                boolean clusterMode = GeneralConfigurationManager.runningInClusterMode();
                List<GPI_KPI_Policy> policies = new ArrayList<GPI_KPI_Policy>();
                Collection<ITComputingContextPolicy> itComputingContextPolicies = eacmInstance.getAllITComputingContextPolicyInstances();
                int servers = 0;
                ServiceCenterServer root = null;
                int clusters = 0;
                int onlineServers = 0;
                int runningTasks = 0;
                for (ApplicationActivity applicationActivity : eacmInstance.getAllApplicationActivityInstances()) {
                    if (applicationActivity.isRunning()) {
                        runningTasks++;
                    }
                }
                //  if (GeneralConfigurationManager.runningInClusterMode()) {
                for (ITComputingContextPolicy policy : itComputingContextPolicies) {
                    ServiceCenterServer serviceCenterServer = (ServiceCenterServer) policy.getPolicySubject().iterator().next();
                    if (clusterMode) {
                        if (serviceCenterServer.isRootCluster()) {
                            policies.add(policy);
                            root = serviceCenterServer;
                        }
                    } else {
                        policies.add(policy);
                    }
                    if (!serviceCenterServer.isCluster()) {
                        servers++;
                        if (serviceCenterServer.getIsActive()) {
                            onlineServers++;
                        }
                    } else {
                        clusters++;
                    }
                }
                //  } else {
                //       servers = eacmInstance.getAllServiceCenterServerInstances().size();
                //      policies.addAll(itComputingContextPolicies);
                //      for (ServiceCenterServer serviceCenterServer : eacmInstance.getAllServiceCenterServerInstances()) {
                ////          if (serviceCenterServer.getIsActive()) {
                //              onlineServers++;
                //  }
                //     }
                // }

                policies.addAll(eacmInstance.getAllQoSPolicyInstances());

                Date before = new Date();


                RLNode initialContext = new RLNode(new LinkedList());
                Pair<Double, List<GPI_KPI_Policy>> entropyAndPolicy = entropyEvaluator.computeEntropy(policies);
                initialContext.setContextEntropy(entropyAndPolicy.getFirst());
                initialContext.setRewardFunction(entropyEvaluator.computeRewardFunction(null, initialContext, null));

                Queue<ContextAction> actionPlanActions = new LinkedList<ContextAction>();

                RLNode rlNodeFromMemory = null;//findSimilarActions(eacmInstance, initialContext, policies);
//                if (rlNodeFromMemory != null) {
//                    actionPlanActions.addAll(rlNodeFromMemory.getActions());
//                } else {
                ActionPlan plan = planningEngine.generateActionPlan(policies);
                actionPlanActions = plan.getActions();
                if (!GeneralConfigurationManager.getMemoryState().equalsIgnoreCase("disabled")) {

                    if ((actionPlanActions.size() > 0) && (plan.getEntropy() == 0)) {

                        ContextSituationDto situationDto = new ContextSituationDto();
                        PowerConsumption powerConsumption = null;
                        try {
                            powerConsumption = new PowerConsumptionEstimator(eacmInstance);
                        } catch (RemoteException e) {
                            GlobalLoopLogger.error(e);
                        }

                        try {
                            situationDto.setEnergyLevel(powerConsumption.getValueWithRunningAlgorithm().doubleValue());
                        } catch (Exception e) {
                            situationDto.setEnergyLevel(1000000000);
                            GlobalLoopLogger.getLogger().debug(e.toString());
                        }
                        Collection<ServiceCenterServer> serversList = eacmInstance.getAllServiceCenterServerInstances();
                        Collection<ApplicationActivity> actionPlanTasks = eacmInstance.getAllApplicationActivityInstances();

                        for (ServiceCenterServer server : serversList) {
                            situationDto.addServerDto(ServerToDtoMapper.map(server));
                        }

                        for (ApplicationActivity activity : actionPlanTasks) {
                            situationDto.addTaskDto(TaskToDtoMapper.map(activity));
                        }
                        for (ContextAction action : actionPlanActions) {
                            situationDto.addActionDto(ActionToDtoMapper.map(action));
                        }


//                        try {
//                            capInteractions.persistContextSituation(situationDto);
//                        } catch (ApplicationException e) {
//                            GlobalLoopLogger.getLogger().debug(e.toString());
//                        }
                        SimilarContextSituationIdentification.getIdentification().addSituationToEnergyClusters(situationDto);
                    }
                }

//                }


                Date after = new Date();

                Date time = new Date(after.getTime() - before.getTime());
                int toBeDeployed = 0;
                for (ApplicationActivity applicationActivity : eacmInstance.getAllApplicationActivityInstances()) {
                    if (!applicationActivity.isRunning()) {
                        toBeDeployed++;
                    }
                }

                ArrayList<String> logMessage = new ArrayList<String>();
                logMessage.add("" + (time.getHours() - 2));
                logMessage.add("" + time.getMinutes());
                logMessage.add("" + time.getSeconds());
                logMessage.add("" + servers);
                logMessage.add("" + clusters);
                logMessage.add("" + toBeDeployed);
                logMessage.add("" + runningTasks);
                logMessage.add("" + onlineServers);
                logMessage.add("" + "" + dateFormat.format(new Date()));
//                logMessage.addAll(getServersLoad());
                String[] splitMessage = new String[logMessage.size()];
                splitMessage = logMessage.toArray(splitMessage);
                simpleCSVWriter.log(splitMessage);

                List<String> generatedActionsNames = new ArrayList<String>(
                        actionPlanActions.size());

//                GlobalLoopLogger.getLogger().info("Generated action plan:");
                for (ContextAction contextAction : actionPlanActions) {
                    generatedActionsNames.add(contextAction.getName());
//                    GlobalLoopLogger.getLogger().info(contextAction);
                }
                if (actionPlanActions.size() == 0) {
                    printServersLoad(root, 0);
                }
                GlobalLoopLogger.getLogger().debug("Final entropy" + plan.getEntropy());
                ActionPlanDto actionPlanDto = new ActionPlanDto(generatedActionsNames);

                GlobalLoopLogger.getLogger().info("");
                InterAgentMessageBuilder.sendMessage(agent, ActionPlanEnforcementAgent.class,
                        AgentMessageType.START_ACTION_PLAN_EXECUTION, actionPlanDto);
            }
        };
        currentlyRunningThread.start();
    }

    public void terminate() {
        if (currentlyRunningThread != null && currentlyRunningThread.isAlive()) {
            currentlyRunningThread.interrupt();
            planningEngine.terminate();
            try {
                currentlyRunningThread.join();
            } catch (InterruptedException e) {
                GlobalLoopLogger.error(e);
            }
        }
    }

    public ArrayList<String> getServersLoad() {
        DecimalFormat decimalFormat = new DecimalFormat("#.##");
        ArrayList<String> loads = new ArrayList<String>();
        for (ServiceCenterITComputingResource resource : eacmInstance.getAllServiceCenterServerInstances()) {
            if (!((ServiceCenterServer) resource).isCluster()) {
                loads.add("" + decimalFormat.format(getActiveServersPercentage((ServiceCenterServer) resource)));
            }
        }
        return loads;
    }

    public void printServersLoad(ServiceCenterServer serviceCenterServers, int index) {
        DecimalFormat decimalFormat = new DecimalFormat("#.##");
        if (serviceCenterServers == null) {
            return;
        }
        for (ServiceCenterITComputingResource resource : serviceCenterServers.getResources()) {
            if (!((ServiceCenterServer) resource).getIsActive()) {
                continue;
            }
            String value = "";
            for (int i = 0; i < index; i++) {
                value += "   ";
            }
            value += resource.getName().split("#")[1] + ":  " + (decimalFormat.format(getActiveServersPercentage((ServiceCenterServer) resource)) + "%");
            System.out.println(value);
            if (((ServiceCenterServer) resource).isCluster()) {
                printServersLoad((ServiceCenterServer) resource, new Integer(index + 1));
            }
        }

    }

    public Double getActiveServersPercentage(ServiceCenterServer serviceCenterServers) {

        Double load = null;

        if (serviceCenterServers.isCluster()) {
            Collection<ServiceCenterITComputingResource> serviceCenterITComputingResources = serviceCenterServers.getResources();
            int resourcesCount = serviceCenterITComputingResources.size();
            int activeResources = 0;
            for (ServiceCenterITComputingResource resource : serviceCenterITComputingResources) {
                if (getActiveServersPercentage((ServiceCenterServer) resource) > 0) {
                    activeResources++;
                }
            }
            load = (activeResources > 0) ? activeResources * 100 / resourcesCount : 0.0;
        } else {
            if (serviceCenterServers.getIsActive()) {
                double cpuLoad = computeCpuUsagePercentage(serviceCenterServers);
                double memLoad = computeMemoryUsagePercentage(serviceCenterServers);
                load = (cpuLoad > memLoad) ? cpuLoad : memLoad;
            } else {
                load = 0.0;
            }
        }
        return load;

    }

    private Double computeCpuUsagePercentage(ServiceCenterServer serviceCenterServer) {
        if (serviceCenterServer.isCluster()) {
            if (serviceCenterServer.getIsActive()) {
                return 100.0;
            } else {
                return 0.0;
            }
        } else {
            CPU cpu = serviceCenterServer.getCpuResources().iterator().next();
            return (cpu.getCurrentWorkLoad() * 100) / cpu.getMaximumWorkLoad();
        }
    }

    private Double computeMemoryUsagePercentage(ServiceCenterServer serviceCenterServer) {
        if (serviceCenterServer.isCluster()) {
            if (serviceCenterServer.getIsActive()) {
                return 100.0;
            } else {
                return 0.0;
            }
        } else {
            MEM mem = serviceCenterServer.getMemResources().iterator().next();
            return (mem.getCurrentWorkLoad() * 100) / mem.getMaximumWorkLoad();
        }
    }
}
