/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.actionPlanningModule.reasoning.impl;


import ro.utcn.dsrl.actionPlanningModule.reasoning.PlanningEngine;
import ro.utcn.dsrl.actionPlanningModule.utils.actionExecutionSimulation.IActionSimulationUtil;
import ro.utcn.dsrl.actionPlanningModule.utils.actionExecutionSimulation.impl.DefaultActionSimulationUtil;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.api.EACMInstance;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.*;

import java.util.*;
import ro.utcn.dsrl.actionPlanningModule.utils.ActionPlan;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.ContextEntropyEvaluator;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */
public class FFDPlanningEngine implements PlanningEngine {
    private EACMInstance eacmInstance;
    private IActionSimulationUtil simulationUtil;
ContextEntropyEvaluator entropyEvaluator;
    public FFDPlanningEngine(ContextEntropyEvaluator entropyEvaluator, EACMInstance eacmInstance) {
        super();
        this.entropyEvaluator = entropyEvaluator;
        this.eacmInstance = eacmInstance;
        simulationUtil = new DefaultActionSimulationUtil(this);

    }

    public Collection<ServiceCenterServer> orderServers(ArrayList<ServiceCenterServer> servers) {
        ArrayList<ServiceCenterServer> orderedServers = new ArrayList<ServiceCenterServer>();
        for (int i = 0; i < servers.size(); i++) {
            orderedServers.add(servers.get(i));
        }
        for (int i = 0; i < orderedServers.size(); i++) {
            for (int j = i + 1; j < orderedServers.size(); j++) {
                if (compare(orderedServers.get(i), orderedServers.get(j)) == -1) {
                    ServiceCenterServer s = orderedServers.get(i);
                    orderedServers.set(i, orderedServers.get(j));
                    orderedServers.set(j, s);
                }
            }
        }
        return orderedServers;
    }

    private int compare(ServiceCenterServer server1, ServiceCenterServer server2) {
        //compute difference betweeen servers
        double difference1 = (server1.getCpuResources().iterator().next().getMaximumWorkLoad() - server1.getCpuResources().iterator().next().getCurrentWorkLoad() - (server2.getCpuResources().iterator().next().getMaximumWorkLoad() - server2.getCpuResources().iterator().next().getCurrentWorkLoad()));
        double difference2 = server1.getMemResources().iterator().next().getMaximumWorkLoad() - server1.getMemResources().iterator().next().getCurrentWorkLoad() - (server2.getMemResources().iterator().next().getMaximumWorkLoad() - server2.getMemResources().iterator().next().getCurrentWorkLoad());
        if (difference1 < 0 && difference2 < 0) return -1;
        if (difference1 > 0 && difference2 > 0) return 1;

        if (difference1 + difference2 < 0) return -1;
        if (difference2 + difference1 > 0) return 1;
        if (difference2 + difference1 == 0) return 0;

        return -23;


    }

    public ActionPlan generateActionPlan(Collection<GPI_KPI_Policy> policies) {
        Queue<ContextAction> contextActions = new LinkedList<ContextAction>();
        ArrayList<ContextAction> actions = new ArrayList<ContextAction>();
        Collection<ServiceCenterServer> servers = new ArrayList<ServiceCenterServer>();
        Collection<ApplicationActivity> activities = new ArrayList<ApplicationActivity>();
        for (GPI_KPI_Policy gpi_kpi_policy : policies) {
            if (gpi_kpi_policy instanceof ITComputingContextPolicy) {

                ServiceCenterServer serviceCenterServer = (ServiceCenterServer) gpi_kpi_policy.getPolicySubject().iterator().next();
                servers.add(serviceCenterServer);

            } else {
                activities.add((ApplicationActivity) gpi_kpi_policy.getPolicySubject().iterator().next());
            }
        }
        Collection<ServiceCenterServer> orderedServers = orderServers((ArrayList<ServiceCenterServer>) servers);

        for (ApplicationActivity activity : activities) {
            //avoid redeploying activities
            if (activity.isRunning()) {
                continue;
            }
            for (ServiceCenterServer server : orderedServers) {

                if (server.hasResourcesFor(activity)) {
                    if (!server.getIsActive()) {
                        SetServerStateActivity newAction=
                                eacmInstance.createSetServerStateActivity("Set_state_for_" + server.getName()
                                        + "_to_" + 1);

                                 newAction.setTargetServerState(1);
                                newAction.addResource(server);
                        simulationUtil.simulateActionExecution(eacmInstance, newAction);
                        contextActions.add(newAction);
                        actions.add(newAction);
                    }
                    DeployActivity deployActivity = eacmInstance.createDeployActivity("Deploy_" + activity.getName() + "_to_" + server.getName());
                    deployActivity.setActivity(activity);
                    deployActivity.setResourceTo(server);
                    simulationUtil.simulateActionExecution(eacmInstance, deployActivity);
                    contextActions.add(deployActivity);
                    actions.add(deployActivity);
                    break;
                }
            }
        }
        for (ServiceCenterServer server : orderedServers) {
            if (!server.hasRunningActivities() && server.getIsActive()) {
                SetServerStateActivity newAction =
                                eacmInstance.createSetServerStateActivity("Set_state_for_" + server.getName()
                                        + "_to_" + 0);
                newAction.setTargetServerState(0);
                newAction.addResource(server);
                simulationUtil.simulateActionExecution(eacmInstance, newAction);
                contextActions.add(newAction);
                actions.add(newAction);

            }
        }
        double entropy = entropyEvaluator.computeEntropy(policies).getFirst();
        for (int i = actions.size() - 1; i >= 0; i--) {
            actions.get(i).undo(eacmInstance);
        }
        return new ActionPlan(contextActions,entropy);
    }

    public void terminate() {

    }
}
