/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.actionPlanningModule.reasoning.impl;

import org.apache.log4j.Logger;
import ro.utcn.dsrl.actionPlanningModule.reasoning.PlanningEngine;
import ro.utcn.dsrl.actionPlanningModule.reasoning.negotiation.Negotiator;
import ro.utcn.dsrl.actionPlanningModule.reasoning.negotiation.impl.NegotiatorFactory;
import ro.utcn.dsrl.actionPlanningModule.utils.actionExecutionSimulation.IActionSimulationUtil;
import ro.utcn.dsrl.actionPlanningModule.utils.actionExecutionSimulation.impl.DefaultActionSimulationUtil;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.api.EACMInstance;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.*;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.ContextEntropyEvaluator;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.RLNode;
import ro.utcn.dsrl.utils.config.GeneralConfigurationManager;
import ro.utcn.dsrl.utils.logging.GlobalLoopLogger;
import ro.utcn.dsrl.utils.misc.Pair;

import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicLong;
import ro.utcn.dsrl.actionPlanningModule.utils.ActionPlan;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */
public class ReinforcementLearningPlanningEngine implements PlanningEngine {

    private Logger logger;
    private Thread planningThread;
    private List<ContextAction> triedActions;
    private static final Object lock = new Object();
    private static final int MAXIMUM_STACK_DEPTH = 100;
    private static final float TOLERABLE_PERCENTAGE = 5;
    private boolean remembered = false;
    //    private SimilarContextSituationIdentification similarContextSituationIdentification;
    private ContextEntropyEvaluator entropyEvaluator;
    private EACMInstance eacmInstance;
    private IActionSimulationUtil simulationUtil;

    public ReinforcementLearningPlanningEngine(ContextEntropyEvaluator entropyEvaluator, EACMInstance eacmInstance) {
        super();
        triedActions = new ArrayList<ContextAction>();
        logger = GlobalLoopLogger.getLogger();
        this.eacmInstance = eacmInstance;
        this.entropyEvaluator = entropyEvaluator;
        simulationUtil = new DefaultActionSimulationUtil(this);
    }

    public ActionPlan generateActionPlan(
            final Collection<GPI_KPI_Policy> policies) {


        final Queue<ContextAction> actions = new LinkedList<ContextAction>();

        System.out.println("Start searching for actions ... ");
        RLNode rlNode = iterativeReinforcementLearning(policies);
        actions.addAll(rlNode.getActions());


        if (entropyEvaluator.computeEntropy(policies).getFirst() > 0) {
            Queue<ContextAction> negotiatedActions = negotiateAndGenerateActionPlan(policies);
            for (ContextAction contextAction : negotiatedActions) {
                actions.add(contextAction);
            }
        }

        if (actions.size() == 0) {
            GlobalLoopLogger.getLogger().info("No solution found. ");
        }

        //try to free up the EACM model from simulated actions which are not used
//        for (ContextAction consolidationAction : triedActions) {
//            if (!actions.contains(consolidationAction)) {
//                consolidationAction.delete();
//            }
//        }
        triedActions.clear();

        return new ActionPlan(actions, rlNode.getContextEntropy());

    }

    private Queue<ContextAction> negotiateAndGenerateActionPlan(final Collection<GPI_KPI_Policy> policies) {

        ArrayList<String> negotiationMessage = new ArrayList<String>();

        List<GPI_KPI_Policy> violatedPolicies = entropyEvaluator.computeEntropy(policies).getSecond();
        List<GPI_KPI_Policy> violatedPoliciesRegardingTasks = new ArrayList<GPI_KPI_Policy>();
        for (GPI_KPI_Policy gpi_kpi_policy : violatedPolicies) {
            if (gpi_kpi_policy.getPolicySubject() instanceof ApplicationActivity) {
                violatedPoliciesRegardingTasks.add(gpi_kpi_policy);
            }
        }
        Collection<ServiceCenterServer> servers = eacmInstance.getAllServiceCenterServerInstances();
        Queue<ContextAction> actionQueue = new PriorityQueue<ContextAction>();
        if (violatedPoliciesRegardingTasks.size() > 0) {
            for (int i = 0; i < violatedPoliciesRegardingTasks.size(); i++) {
                if (violatedPoliciesRegardingTasks.get(i).getPolicySubject().get(0) instanceof ApplicationActivity) {

                    ApplicationActivity activity = (ApplicationActivity) violatedPoliciesRegardingTasks.get(0).getPolicySubject().get(0);

                    Negotiator negotiator = NegotiatorFactory.getNashNegotiator();
                    Collection<ServiceCenterServer> sortedServers = sortServersByDistanceToActivity(servers, activity);
                    if (sortedServers.size() > 0) {
                        ServiceCenterServer server = sortedServers.iterator().next();
                        if (server != null) {
                            negotiationMessage.add("Negotiating between " + activity.getLocalName() + " and " + server.getLocalName());
                            if (server != null) {
                                CPU cpu = server.getCpuResources().iterator().next();
                                MEM mem = server.getMemResources().iterator().next();
                                Map<String, Double> negotiatedValues = negotiator.negotiate(server, activity);
                                double optimalCpu = cpu.getOptimalWorkLoad();
                                double totalCpu = cpu.getMaximumWorkLoad();
                                double currentCpu = cpu.getCurrentWorkLoad();
                                double optimalMem = mem.getOptimalWorkLoad();

                                if (negotiatedValues.get(Negotiator.NEGOTIATED_CPU) + currentCpu > (optimalCpu + totalCpu) / 2.0) {
                                    double current = negotiatedValues.get(Negotiator.NEGOTIATED_CPU) + currentCpu;
                                    optimalCpu = 2 * current - cpu.getMaximumWorkLoad();
                                }
                                if (negotiatedValues.get(Negotiator.NEGOTIATED_MEMORY) + mem.getCurrentWorkLoad() > (mem.getOptimalWorkLoad() + mem.getMaximumWorkLoad()) / 2.0) {
                                    double current = negotiatedValues.get(Negotiator.NEGOTIATED_MEMORY) + mem.getCurrentWorkLoad();
                                    optimalMem = 2 * current - mem.getMaximumWorkLoad();
                                }
                                if (!server.getIsActive()) {
                                    SetServerStateActivity newActivity =
                                            eacmInstance.createSetServerStateActivity("Set_state_for_" + server.getName()
                                            + "_to_" + 1);

                                    newActivity.setTargetServerState(1);
                                    newActivity.addResource(server);
                                    actionQueue.add(newActivity);
                                }

                                ServerAdaptationAction defaultServerAdaptationAction = eacmInstance.createServerAdaptationAction("ServerAdaptationAction_" + server);
                                defaultServerAdaptationAction.setNewOptimalValueForCpu((int) optimalCpu);
                                defaultServerAdaptationAction.setNewOptimalValueForMem((int) optimalMem);
                                defaultServerAdaptationAction.setServer(server);
                                negotiationMessage.add(defaultServerAdaptationAction.toString());
                                actionQueue.add(defaultServerAdaptationAction);

                                ApplicationAdaptationAction applicationAdaptationAction = eacmInstance.createApplicationAdaptationAction("ApplicationAdaptationAction_" + activity.getLocalName());
                                applicationAdaptationAction.setActivity(activity);
                                applicationAdaptationAction.setCpuMin((int) activity.getCpuRequiredMinValue());
                                applicationAdaptationAction.setCpuMax(negotiatedValues.get(Negotiator.NEGOTIATED_CPU).intValue());
                                applicationAdaptationAction.setMemMin((int) activity.getMemRequiredMinValue());
                                applicationAdaptationAction.setMemMax(negotiatedValues.get(Negotiator.NEGOTIATED_MEMORY).intValue());

                                negotiationMessage.add(applicationAdaptationAction.toString());

                                DeployActivity deployActivityAction = eacmInstance.createDeployActivity("Deploy_"
                                        + activity.getName() + "_to_" + server.getName());

                                deployActivityAction.setActivity(activity);
                                deployActivityAction.setResourceTo(server);
                                actionQueue.add(deployActivityAction);


                            }
                        }
                    }

                }
            }
        }

        return actionQueue;
    }

    private RLNode iterativeReinforcementLearning(
            Collection<GPI_KPI_Policy> policiesForEvaluation) {
        int stackDepth = 0;
        int index = 0;
        RLNode newContext = null;
        PriorityQueue<RLNode> queue = new PriorityQueue<RLNode>();
        RLNode smallestEntropyContext = new RLNode(new PriorityQueue<ContextAction>());

        Pair<Double, List<GPI_KPI_Policy>> initialContextState = entropyEvaluator.computeEntropy(policiesForEvaluation);

        smallestEntropyContext.setContextEntropy(initialContextState.getFirst());
        smallestEntropyContext.setBrokenPolicies(initialContextState.getSecond());

        queue.add(smallestEntropyContext);
        ServiceCenterServer root = null;
        Collection<ServiceCenterServer> servers = new ArrayList<ServiceCenterServer>();

        for (GPI_KPI_Policy gpi_kpi_policy : policiesForEvaluation) {
            if (gpi_kpi_policy instanceof ITComputingContextPolicy) {
                ServiceCenterServer serviceCenterServer = (ServiceCenterServer) gpi_kpi_policy.getPolicySubject().iterator().next();

                servers.add(serviceCenterServer);
                if (serviceCenterServer.isRootCluster()) {
                    root = serviceCenterServer;
                }
            }
        }

        while ((queue.peek() != null) && stackDepth <= MAXIMUM_STACK_DEPTH) {

            if (planningThread != null && planningThread.isInterrupted()) {
                return new RLNode(new PriorityQueue<ContextAction>());
//                planningThread.stop();
            }

            newContext = queue.poll();

            //create  map of actions used in current context to avoid cycles
            Queue<ContextAction> currentContextActionsQueue = newContext.getActions();

            //to check easily if we enter cycles : turn same server ON, OFF, ON, OFF and so on
            TreeSet<SetServerStateActivity> turnOnOffActions = new TreeSet<SetServerStateActivity>(new Comparator<SetServerStateActivity>() {

                public int compare(SetServerStateActivity o1, SetServerStateActivity o2) {

                    String name1 = o1.getName();
                    String name2 = o2.getName();
                    if (name1.equals(name2) || o1.getResources().equals(o2.getResources())) {
                        return 0;
                    } else {
                        return name1.compareTo(name2);
                    }
                }
            });

            TreeSet<ConsolidationAction> migrateAndDeployActions = new TreeSet<ConsolidationAction>(new Comparator<ConsolidationAction>() {

                public int compare(ConsolidationAction o1, ConsolidationAction o2) {

                    String name1 = o1.getName();
                    String name2 = o2.getName();

                    if (o1 instanceof MigrateActivity) {
                        MigrateActivity activityAction1 = (MigrateActivity) o1;

                        if (o2 instanceof MigrateActivity) {
                            MigrateActivity activityAction2 = (MigrateActivity) o2;

                            if (activityAction1.getActivity().equals(activityAction2.getActivity())) {
                                return 0;
                            } else {
                                return name1.compareTo(name2);
                            }
                        } else if (o2 instanceof DeployActivity) {
                            DeployActivity activityAction2 = (DeployActivity) o2;

                            if (activityAction1.getActivity().equals(activityAction2.getActivity())) {
                                return 0;
                            } else {
                                return name1.compareTo(name2);
                            }
                        }
                    } else if (o1 instanceof DeployActivity) {
                        DeployActivity activityAction1 = (DeployActivity) o1;

                        if (o2 instanceof MigrateActivity) {

                            MigrateActivity activityAction2 = (MigrateActivity) o2;
                            if (activityAction1.getActivity().equals(activityAction2.getActivity())) {
                                return 0;
                            } else {
                                return name1.compareTo(name2);
                            }
                        } else if (o2 instanceof DeployActivity) {
                            DeployActivity activityAction2 = (DeployActivity) o2;
                            if (activityAction1.getActivity().equals(activityAction2.getActivity())) {
                                return 0;
                            } else {
                                return name1.compareTo(name2);
                            }
                        }
                    }
                    return name1.compareTo(name2);
                }
            });


            for (ContextAction action : currentContextActionsQueue) {
                if (action instanceof SetServerStateActivity) {
                    turnOnOffActions.add((SetServerStateActivity) action);
                } else if (action instanceof ConsolidationAction) {
                    migrateAndDeployActions.add((ConsolidationAction) action);
                }
            }


            remembered = false;

            newContext.executeActions(eacmInstance, simulationUtil);
            Pair<Double, List<GPI_KPI_Policy>> entropyAndPolicy = entropyEvaluator.computeEntropy(policiesForEvaluation);
            //trying to deploy all tasks. To get maximum QoS While still turning off servers


            if (smallestEntropyContext.compareTo(newContext) > 0) {
                smallestEntropyContext = newContext;
            }
            List<ApplicationActivity> associatedTasks = new ArrayList<ApplicationActivity>();
            List<ServiceCenterServer> associatedServers = new ArrayList<ServiceCenterServer>();


            if (entropyAndPolicy.getFirst() > 0) {
                if (entropyAndPolicy.getSecond().size() > 0) {
                    List<GPI_KPI_Policy> policies = entropyAndPolicy.getSecond();

                    for (GPI_KPI_Policy policy : policies) {
                        if (policy.getPolicySubject().get(0) instanceof ApplicationActivity) {
                            associatedTasks.add((ApplicationActivity) policy.getPolicySubject().get(0));
                        } else {
                            ServiceCenterServer associatedServer = (ServiceCenterServer) policy.getPolicySubject().get(0);
                            associatedServers.add(associatedServer);
                        }
                    }
                }
                HashMap<ApplicationActivity, Collection<ServiceCenterServer>> applicationDictionary = new HashMap<ApplicationActivity, Collection<ServiceCenterServer>>();
                for (ApplicationActivity activity : associatedTasks) {
                    applicationDictionary.put(activity, sortServersByDistanceToActivity(servers, activity));
                }
//                if (stackDepth >= MAXIMUM_STACK_DEPTH) {
//                    if (associatedTasks.size() <= 0) {
//                        break;
//                    } else {
//                if (associatedTasks.size() > 1) {
//                    logger.info(simulationResultMessage);
//                    for (ContextAction action : actions) {
//                        logger.info("" + action.toString());
//                    }
//                    logger.info("QUEUE");
//                    for(RLNode cs: queue){
//                        logger.info("\n " +cs.toString());
//                        for(ContextAction ca : cs.getActions()){
//                            logger.info(ca.toString());
//                        }
//                    }
//                }
//                }

                if (associatedTasks.size() > 0) {
                    // deploy actions
                    for (ApplicationActivity activity : associatedTasks) {

                        for (ServiceCenterServer serverInstance : applicationDictionary.get(activity)) {

                            if (serverInstance.getIsActive()
                                    && !activity.isRunning()
                                    && serverInstance.hasResourcesFor(activity)
                                    && !serverInstance.hostsActivity(activity)) {

                                DeployActivity newAction = eacmInstance.createDeployActivity("Deploy_" + activity.getName() + "_to_" + serverInstance.getName());
                                newAction.setResourceTo(serverInstance);
                                newAction.setActivity(activity);
                                triedActions.add(newAction);

                                if (!migrateAndDeployActions.contains(newAction)) {

                                    RLNode currentRLNode = new RLNode(new LinkedList(newContext.getActions()));
                                    currentRLNode.getActions().add(newAction);

                                    synchronized (lock) {
                                        logger.debug("Simulating " + "Deploy_" + activity.getName() + "_to_" + serverInstance.getName());
                                        simulationUtil.simulateActionExecution(eacmInstance, newAction);
                                        Pair<Double, List<GPI_KPI_Policy>> afterExecuteEntropy = entropyEvaluator.computeEntropy(policiesForEvaluation);
                                        currentRLNode.setContextEntropy(afterExecuteEntropy.getFirst());
                                        currentRLNode.setBrokenPolicies(afterExecuteEntropy.getSecond());
                                        currentRLNode.setRewardFunction(entropyEvaluator.computeRewardFunction(
                                                newContext, currentRLNode,
                                                new LinkedList(currentRLNode.getActions())));
                                        simulationUtil.undoActionSimulatedExecution(eacmInstance, newAction);

                                        if (afterExecuteEntropy.getFirst() == 0) {
                                            newContext.undoAllActions(eacmInstance);
                                            return currentRLNode;
                                        }
                                    }
                                    if (!queue.contains(currentRLNode)) {
                                        if (serverInstance.isCluster()) {
                                            if (newAction.hasExecutedSubActions()) {
                                                queue.add(currentRLNode);
                                            }
                                        } else {
                                            queue.add(currentRLNode);
                                        }
                                    }
                                }

                            } else if (!serverInstance.getIsActive()
                                    && serverInstance.hasResourcesFor(activity)) {
                                SetServerStateActivity newAction = eacmInstance.createSetServerStateActivity("Set_state_for_"
                                        + serverInstance.getName()
                                        + "_to_" + 1);
                                RLNode currentRLNode = new RLNode(new LinkedList<ContextAction>(newContext.getActions()));
                                newAction.setTargetServerState(1);
                                newAction.addResource(serverInstance);
                                triedActions.add(newAction);

                                // if action is not already in the actions list
                                if (!turnOnOffActions.contains(newAction)) {
                                    currentRLNode.getActions().add(newAction);

                                    synchronized (this) {
                                        simulationUtil.simulateActionExecution(eacmInstance, newAction);
                                        Pair<Double, List<GPI_KPI_Policy>> afterExecuteEntropy = entropyEvaluator.computeEntropy(policiesForEvaluation);
                                        currentRLNode.setContextEntropy(afterExecuteEntropy.getFirst());
                                        currentRLNode.setBrokenPolicies(afterExecuteEntropy.getSecond());

                                        currentRLNode.setRewardFunction(entropyEvaluator.computeRewardFunction(
                                                newContext,
                                                currentRLNode,
                                                new LinkedList<ContextAction>(currentRLNode.getActions())));
                                        simulationUtil.undoActionSimulatedExecution(eacmInstance, newAction);
                                        if (afterExecuteEntropy.getFirst() == 0) {
                                            newContext.undoAllActions(eacmInstance);
                                            return currentRLNode;
                                        }

                                    }

                                    if (!queue.contains(currentRLNode)) {
                                        queue.add(currentRLNode);
                                    }

                                }
                            }
                        }
                    }
                }

//                move actions

                for (ServiceCenterServer sourceServer : servers) {
                    if (GeneralConfigurationManager.isMigrationEnabled()) {

                        if (!sourceServer.isCluster() && sourceServer.getIsActive()) {
                            for (ApplicationActivity applicationActivity : sourceServer.getRunningActivities()) {
                                //TODO: de folosit dictionaru de distante
                                for (ServiceCenterServer destinationServer : sortServersByDistanceToActivity(servers, applicationActivity)) {

                                    //migrate only among atomic ServiceCenterServer entities, not between clusters when in cluster mode
                                    if (!destinationServer.isCluster() && destinationServer.getIsActive()
                                            && destinationServer.hasResourcesFor(applicationActivity)
                                            && !destinationServer.hostsActivity(applicationActivity)) {
                                        // if (!migrationGeneratesCycle(newContext,
                                        // applicationActivity)) {
                                        MigrateActivity newAction = eacmInstance.createMigrateActivity("Migrate_from_"
                                                + destinationServer.getName()
                                                + "_to_"
                                                + applicationActivity.getName()
                                                + "_Activity");
                                        triedActions.add(newAction);

                                        RLNode currentRLNode = new RLNode(new LinkedList(newContext.getActions()));
                                        newAction.setResourceFrom(sourceServer);
                                        newAction.setResourceTo(destinationServer);
                                        newAction.setActivity(applicationActivity);

                                        if (!migrateAndDeployActions.contains(newAction)) {

                                            currentRLNode.getActions().add(newAction);

                                            synchronized (lock) {
                                                simulationUtil.simulateActionExecution(eacmInstance, newAction);

                                                Pair<Double, List<GPI_KPI_Policy>> afterExecuteEntropy = entropyEvaluator.computeEntropy(policiesForEvaluation);
                                                currentRLNode.setContextEntropy(afterExecuteEntropy.getFirst());
                                                currentRLNode.setBrokenPolicies(afterExecuteEntropy.getSecond());

                                                currentRLNode.setRewardFunction(entropyEvaluator.computeRewardFunction(
                                                        newContext, currentRLNode,
                                                        new LinkedList<ContextAction>(
                                                        currentRLNode.getActions())));

                                                simulationUtil.undoActionSimulatedExecution(eacmInstance, newAction);
                                                if (afterExecuteEntropy.getFirst() == 0) {
                                                    newContext.undoAllActions(eacmInstance);
                                                    return currentRLNode;
                                                }
                                            }
                                            if (!queue.contains(currentRLNode)) {
                                                if (destinationServer.isCluster() || sourceServer.isCluster()) {
                                                    if (newAction.hasExecutedSubActions()) {
                                                        queue.add(currentRLNode);
                                                    }
                                                } else {
                                                    queue.add(currentRLNode);
                                                }
                                            }
                                        }
                                    } else {
                                        if (!destinationServer.getIsActive()
                                                && destinationServer.hasResourcesFor(applicationActivity)) {
                                            SetServerStateActivity newAction = eacmInstance.createSetServerStateActivity("Set_state_for_"
                                                    + destinationServer.getName()
                                                    + "_to_" + 1);
                                            triedActions.add(newAction);
                                            RLNode currentRLNode = new RLNode(
                                                    new LinkedList(newContext.getActions()));
                                            newAction.setTargetServerState(1);
                                            newAction.addResource(destinationServer);

                                            if (!turnOnOffActions.contains(newAction)) {
                                                currentRLNode.getActions().add(newAction);

                                                synchronized (lock) {
                                                    simulationUtil.simulateActionExecution(eacmInstance, newAction);
                                                    Pair<Double, List<GPI_KPI_Policy>> afterExecuteEntropy = entropyEvaluator.computeEntropy(policiesForEvaluation);
                                                    currentRLNode.setContextEntropy(afterExecuteEntropy.getFirst());
                                                    currentRLNode.setBrokenPolicies(afterExecuteEntropy.getSecond());

                                                    currentRLNode.setRewardFunction(entropyEvaluator.computeRewardFunction(
                                                            newContext,
                                                            currentRLNode,
                                                            new LinkedList<ContextAction>(
                                                            currentRLNode.getActions())));

                                                    simulationUtil.undoActionSimulatedExecution(eacmInstance, newAction);
                                                    if (afterExecuteEntropy.getFirst() == 0) {
                                                        newContext.undoAllActions(eacmInstance);
                                                        return currentRLNode;
                                                    }
                                                }
                                                if (!queue.contains(currentRLNode)) {
                                                    queue.add(currentRLNode);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }

                    //turn off unused servers
                    if (sourceServer.getIsActive()
                            && (sourceServer.getRunningActivities().size() == 0)) {
                        SetServerStateActivity newAction = eacmInstance.createSetServerStateActivity("Set_state_for_"
                                + sourceServer.getName() + "_to_" + 0);
                        RLNode currentRLNode = new RLNode(new LinkedList(newContext.getActions()));
                        newAction.setTargetServerState(0);
                        newAction.addResource(sourceServer);
                        triedActions.add(newAction);
                        if (!turnOnOffActions.contains(newAction)) {

                            currentRLNode.getActions().add(newAction);
                            synchronized (lock) {

                                //logger.info("Simulating " + newAction);
                                simulationUtil.simulateActionExecution(eacmInstance, newAction);
                                Pair<Double, List<GPI_KPI_Policy>> afterExecuteEntropy = entropyEvaluator.computeEntropy(policiesForEvaluation);
                                currentRLNode.setContextEntropy(afterExecuteEntropy.getFirst());
                                currentRLNode.setBrokenPolicies(afterExecuteEntropy.getSecond());

                                currentRLNode.setRewardFunction(entropyEvaluator.computeRewardFunction(
                                        newContext,
                                        currentRLNode,
                                        new LinkedList<ContextAction>(currentRLNode.getActions())));

                                simulationUtil.undoActionSimulatedExecution(eacmInstance, newAction);

                                if (afterExecuteEntropy.getFirst() == 0) {
                                    newContext.undoAllActions(eacmInstance);
                                    return currentRLNode;
                                }
                            }
                            if (!queue.contains(currentRLNode)) {
                                queue.add(currentRLNode);
                            }
                        }
                    }
                }

                newContext.undoAllActions(eacmInstance);
            } else {
                logger.info("Found solution");
                for (ContextAction a : newContext.getActions()) {
                    logger.info("Solution: " + a.toString());
                }
                newContext.undoAllActions(eacmInstance);
                return newContext;
            }
            stackDepth++;
        }

        Queue<ContextAction> commands = smallestEntropyContext.getActions();
        if (commands.size() == 0) {
            GlobalLoopLogger.getLogger().debug(" Stack depth reached: " + stackDepth);
        }
        return smallestEntropyContext;
    }

    private Collection<ServiceCenterServer> sortServersByDistanceToActivity(
            Collection<ServiceCenterServer> servers, ApplicationActivity task) {
        Map<Double, ServiceCenterServer> bestServersInOrder = new TreeMap<Double, ServiceCenterServer>();


        for (ServiceCenterServer server : servers) {
            Double distance = getClusterDistanceToActivity(server, task);
            //TODO: aici putem controla daca vrem uniformizare uniforma intre deployment clustere sau vrem fiecare cluster umplut pe rand
            //daca las sa suprascrie tot timpu ramane ultimu asa ca o sa distribuie taskurile pe ult cluster that matches tot timpu.
            //aia inseamna ceva ciudat de forma
            /*
            1 0 0  1 0 0
            1 0 0  1 1 0
            1 0 0  1 1 1
            1 1 0  1 1 1
            1 1 1  1 1 1
             */
            if (!bestServersInOrder.containsKey(distance)) {
                bestServersInOrder.put(distance, server);
            }
        }
        return bestServersInOrder.values();
    }

    private Double getClusterDistanceToActivity(
            ServiceCenterServer server, final ApplicationActivity task) {

        long minDif = 10000000;
        final AtomicLong distance = new AtomicLong(minDif);

        if (server.isCluster()) {
            Collection<ServiceCenterITComputingResource> servers = server.getResources();

            final CountDownLatch countDownLatch = new CountDownLatch(servers.size());
            for (final ServiceCenterITComputingResource serviceCenterITComputingResource : servers) {
//                Thread thread = new Thread() {
//                    @Override
//                    public void run() {

                ServiceCenterServer serviceCenterServer = (ServiceCenterServer) serviceCenterITComputingResource;
                synchronized (task) {
                    if (serviceCenterServer.hasResourcesFor(task)) {
                        Double clusterDistance = getClusterDistanceToActivity(serviceCenterServer, task);
                        if (clusterDistance < distance.get()) {
                            distance.set(clusterDistance.longValue());
                        }
                    }
                }
//                        countDownLatch.countDown();
//                    }
//
//                };
//                thread.start();
            }
//            try {
//                countDownLatch.await();
//            } catch (InterruptedException e) {
//                GlobalLoopLogger.error(e);
//            }
        } else {
            double difference = 0.0d;
            double optimal = 0.0;
            double total = 0.0;
            Collection<Core> cores = new ArrayList<Core>();
            for (CPU resource : server.getCpuResources()) {
                difference += Math.pow(
                        (resource.getMaximumWorkLoad() + resource.getOptimalWorkLoad())
                        / 2.0
                        - resource.getCurrentWorkLoad()
                        - task.getCpuRequiredMaxValue(), 2);
                // cores = resource.getAssociatedCores();
                // for (Core core : cores) {
                // optimal = core.getOptimalWorkLoad();
                // total = core.getMaximumWorkLoad();
                // difference += Math.pow(total - (total - optimal) / 2.0 -
                // core.getCurrentWorkLoad()
                // - optimal, 2);
                // }
            }

            for (MEM mem : server.getMemResources()) {
                optimal = mem.getOptimalWorkLoad();
                total = mem.getMaximumWorkLoad();

                difference += Math.pow(
                        total - (total - optimal) / 2.0
                        - mem.getCurrentWorkLoad()
                        - task.getMemRequiredMaxValue() / 2.0, 2);
            }
            for (HDD hdd : server.getHddResources()) {
                optimal = hdd.getOptimalWorkLoad();
                total = hdd.getMaximumWorkLoad();
                difference += Math.pow(
                        total - (total - optimal) / 2.0
                        - hdd.getCurrentWorkLoad()
                        - task.getHddRequiredMaxValue() / 2.0, 2);
            }
            difference = Math.sqrt(difference);
            return difference;
        }
        return distance.doubleValue();
    }

    public void terminate() {
        if (planningThread != null & planningThread.isAlive()) {
            planningThread.interrupt();
            try {
                planningThread.join();
            } catch (InterruptedException e) {
                GlobalLoopLogger.error(e);
            }
        }

    }

    public void printServersLoad(ServiceCenterServer serviceCenterServers, int index) {
        if (serviceCenterServers == null) {
            return;
        }
        for (ServiceCenterITComputingResource resource : serviceCenterServers.getResources()) {
            String value = "";
            for (int i = 0; i < index; i++) {
                value += "   ";
            }
            value += resource.getName().split("#")[1] + ":  " + (getActiveServersPercentage((ServiceCenterServer) resource) + "%");
            System.out.println(value);
            if (((ServiceCenterServer) resource).isCluster()) {
                printServersLoad((ServiceCenterServer) resource, new Integer(index + 1));
            }
        }

    }

    public Double getActiveServersPercentage(ServiceCenterServer serviceCenterServers) {

        Double load = null;

        if (serviceCenterServers.isCluster()) {
            Collection<ServiceCenterITComputingResource> serviceCenterITComputingResources = serviceCenterServers.getResources();
            int resourcesCount = serviceCenterITComputingResources.size();
            int activeResources = 0;
            for (ServiceCenterITComputingResource resource : serviceCenterITComputingResources) {
                if (getActiveServersPercentage((ServiceCenterServer) resource) > 0) {
                    activeResources++;
                }
            }
            load = (activeResources > 0) ? activeResources * 100 / resourcesCount : 0.0;
        } else {
            load = (serviceCenterServers.getIsActive()) ? 100.0 : 0.0;
        }
        return load;

    }
}
