/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.workloadScheduler;

import jade.core.Agent;
import java.io.BufferedReader;
import java.util.logging.Level;
import java.util.logging.Logger;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.CAPInteractions;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.api.EACMInstance;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.ApplicationActivity;
import ro.utcn.dsrl.utils.exposedFunctionality.jaxbBindingClasses.*;
import ro.utcn.dsrl.utils.exposedFunctionality.rmiAPI.RMIAPI;
import ro.utcn.dsrl.utils.exposedFunctionality.rmiAPI.RMICommand;
import ro.utcn.dsrl.utils.commonDtos.*;
import ro.utcn.dsrl.utils.config.GeneralConfigurationManager;
import ro.utcn.dsrl.utils.exceptions.ApplicationException;
import ro.utcn.dsrl.utils.logging.GlobalLoopLogger;
import ro.utcn.dsrl.workloadScheduler.workloadGenerator.FileInputWorkloadGenerator;
import ro.utcn.dsrl.workloadScheduler.workloadGenerator.RandomWorkloadGenerator;
import ro.utcn.dsrl.workloadScheduler.workloadGenerator.WorkloadGeneratorI;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.Marshaller;
import java.io.File;
import java.io.FileReader;
import java.io.Serializable;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.ServiceCenterManagement;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.factories.ServiceCenterManagementFactory;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.factories.impl.DefaultServiceCenterManagementFactory;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.impl.OpenNebulaServiceCenterManagement;
import ro.utcn.dsrl.utils.exceptions.ServiceCenterAccessException;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */
public class WorkloadSchedulerAgent extends Agent {

    private static int BEGINNING_ARRIVAL = 0;
    //interval at which new tasks are generated in milliseconds
    private static int WORKLOAD_GENERATION_INTERVAL_MAX = 12 * 60000;//00;
    private static int WORKLOAD_GENERATION_INTERVAL_MIN = 4 * 60000;//00;
    //the minimum and maximum value for the workload duration(task running time until the task terminates) in seconds
    private static int TASK_DURATION_INTERVAL_MAX = 600;
    private static int TASK_DURATION_INTERVAL_MIN = 320;
    //number of tasks to be generated at one time
    private static int NUMBER_OF_GENERATED_TASKS_MAX = 10;
    private static int NUMBER_OF_GENERATED_TASKS_MIN = 3;
    private Random random = new Random();
    private int generatedTaskIndex = 0;
    private static final DateFormat dateFormat = new SimpleDateFormat("_dd_MM_HH_mm_");
    private Agent agent;
    private boolean pauseGenerating = false;
    private Date agentCreationDate;
    private WorkloadGeneratorI workloadGenerator;
    private File generatedSchedule;
    private WorkloadSchedule workloadSchedule;
    private long lastDelay = 0;
    private EACMInstance eacmInstance;
    private CAPInteractions capInteractions;
    private ServiceCenterManagement serviceCenterManagement;

    @Override
    protected void setup() {

        generatedSchedule = new File("./generatedSchedules/generated_schedule_" + dateFormat.format(new Date()) + ".xml");
        workloadSchedule = new WorkloadSchedule();
        workloadSchedule.setApplications(new JaxbPair[0]);
        Object[] receivedData = getArguments();

        if (receivedData != null && receivedData.length > 0) {
            eacmInstance = (EACMInstance) receivedData[0];
        }
        if (receivedData != null && receivedData.length > 1) {
            capInteractions = (CAPInteractions) receivedData[1];
        }

        workloadGenerator = (GeneralConfigurationManager.isFileBasedWorkloadGenerator())
                ? new FileInputWorkloadGenerator(GeneralConfigurationManager.getWorkloadGeneratorInputFile()) : new RandomWorkloadGenerator(eacmInstance);

        addBehaviour(new ReceiveMessagesWorkloadSchedulerBehavior(this));


        int auxMax = WORKLOAD_GENERATION_INTERVAL_MAX;
        int auxMin = WORKLOAD_GENERATION_INTERVAL_MIN;

        WORKLOAD_GENERATION_INTERVAL_MAX = BEGINNING_ARRIVAL;
        WORKLOAD_GENERATION_INTERVAL_MIN = BEGINNING_ARRIVAL;
        WORKLOAD_GENERATION_INTERVAL_MAX = auxMax;
        WORKLOAD_GENERATION_INTERVAL_MIN = auxMin;
        agentCreationDate = new Date();

        serviceCenterManagement = DefaultServiceCenterManagementFactory.getInstance().createServiceCenterManagement();
        agent = this;

        GlobalLoopLogger.getLogger().info("WSA started");

        generateTasks();

    }

    @Override
    protected void takeDown() {
        pauseGenerating = true;
    }

    public void generateTasks() {

//        if (pauseGenerating) {
//            return;
//        }
////        final Timer timer = new Timer(true);
//
////        TimerTask task = new TimerTask() {
////            int executionCount = 0;
//
////            public void run() {

        long newDelay = random.nextInt(WORKLOAD_GENERATION_INTERVAL_MAX - WORKLOAD_GENERATION_INTERVAL_MIN) + WORKLOAD_GENERATION_INTERVAL_MIN;
        if (workloadSchedule.getApplications().length != 0) {
            lastDelay += newDelay;
        }

        if (!GeneralConfigurationManager.isFileBasedWorkloadGenerator()) {
            try {
                GlobalLoopLogger.getLogger().info("Waiting for schedule for " + newDelay / 1000 + " seconds");
                if (workloadSchedule.getApplications().length != 0) {
                    Thread.sleep(0);
                }

            } catch (InterruptedException e) {
                GlobalLoopLogger.error(e);
            }
        }
        Serializable[] generatedTasks = workloadGenerator.generateTasks();
        List<ActivityDto> activityDtosToBeDeployed = new ArrayList<ActivityDto>();
        List<ActivityDto> activityDtosToBeDeleted = new ArrayList<ActivityDto>();

        boolean thereArePendingVMs = false;

        for (Serializable serializable : generatedTasks) {
            ActivityDto activityDto = (ActivityDto) serializable;
            if (activityDto.getServiceName().contains("deleted")) {
//                try {
//                    List<VirtualTaskInfo> virtualMachines = new OpenNebulaServiceCenterManagement().getAllVirtualMachines();
//                    for (VirtualTaskInfo virtualMachine : virtualMachines){
//                                  System.out.print(virtualMachine.getTaskName());
//
//                       if  (virtualMachine.getTaskName().equals(activityDto.getTaskName())){
//                            activityDtosToBeDeleted.add(activityDto);
//                            //System.out.println( activityDto.getId());
//                       }
//
//                    }
//                } catch (ServiceCenterAccessException e) {
//                    e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
//                }

                //newDelayTODO: uncomment this
                if ((GeneralConfigurationManager.getPlanningEngine().equals("ffd") || GeneralConfigurationManager.getPlanningEngine().equals("rl"))) {
                    ApplicationActivity applicationActivity = eacmInstance.getApplicationActivity(activityDto.getTaskName() + "_" + activityDto.getIp());
                    if (applicationActivity != null && applicationActivity.isRunning()) {
                        activityDtosToBeDeleted.add(activityDto);
                    } else {
                        for (ApplicationActivity aa : eacmInstance.getAllApplicationActivityInstances()) {
                            if (!aa.isRunning()) {
                                thereArePendingVMs = true;
                                break;
                            }
                        }

                    }
                } else {
                    try {
                        ActivityDto deployedService = capInteractions.getDeployedService(activityDto);
//                        for (ActivityDto activity : activityDtos){
                        if (deployedService != null) {
                            activityDtosToBeDeleted.add(deployedService);
//                          }
                        } else {
                            thereArePendingVMs = true;
                        }
                    } catch (ApplicationException e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    }
                }
            } else {
                activityDtosToBeDeployed.add(activityDto);
            }
        }

        Serializable[] activityDtosToBeDeployedSerializable = new Serializable[activityDtosToBeDeployed.size()];
        Serializable[] activityDtosToBeUndeployedSerializable = new Serializable[activityDtosToBeDeleted.size()];

        activityDtosToBeDeployedSerializable = activityDtosToBeDeployed.toArray(activityDtosToBeDeployedSerializable);
        activityDtosToBeUndeployedSerializable = activityDtosToBeDeleted.toArray(activityDtosToBeUndeployedSerializable);

        List<ActivityDto> toBePersisted = new ArrayList<ActivityDto>();
        toBePersisted.addAll(activityDtosToBeDeployed);
        toBePersisted.addAll(activityDtosToBeDeleted);

        GlobalLoopLogger.getLogger().info("To be deployed " + activityDtosToBeDeployed.size());
        GlobalLoopLogger.getLogger().info("To be deleted " + activityDtosToBeDeleted.size());

        if (!GeneralConfigurationManager.isFileBasedWorkloadGenerator()) {
            addActivitiesAndPersistSchedule(toBePersisted, lastDelay);
        }

        boolean somethingWasCalled = false;

        if (!GeneralConfigurationManager.workloadForOpenNebulaHooks()) {
            if (activityDtosToBeUndeployedSerializable.length > 0) {
                try {
                    somethingWasCalled = true;
                    RMIAPI rmiapi;
                    Registry registry = LocateRegistry.getRegistry(GeneralConfigurationManager.getRMIAPIPort());
                    rmiapi = (RMIAPI) registry.lookup(RMIAPI.class.getName());
                    rmiapi.processCommand(RMICommand.SERVICE_COLLECTION_UNDEPLOYMENT_NOTIFICATION, activityDtosToBeUndeployedSerializable);
                } catch (Exception ex) {
                    GlobalLoopLogger.error(ex);
                    ex.printStackTrace();
                }
            }

            //only deploy if there are no pending activities. if pending activities exist it means that the service center is
            //over saturated with activities

            GlobalLoopLogger.getLogger().info("Pending tasks " + thereArePendingVMs);
            if (!thereArePendingVMs && activityDtosToBeDeployedSerializable.length > 0) {
                try {
                    somethingWasCalled = true;
                    RMIAPI rmiapi;
                    Registry registry = LocateRegistry.getRegistry(GeneralConfigurationManager.getRMIAPIPort());
                    rmiapi = (RMIAPI) registry.lookup(RMIAPI.class.getName());
                    rmiapi.processCommand(RMICommand.SERVICE_COLLECTION_DEPLOYMENT_NOTIFICATION, activityDtosToBeDeployedSerializable);
                } catch (Exception ex) {
                    GlobalLoopLogger.error(ex);
                    ex.printStackTrace();
                }
            }
        } else {

            //deploy virtual machines
            if (!thereArePendingVMs && activityDtosToBeDeployedSerializable.length > 0) {
                for (ActivityDto dto : activityDtosToBeDeployed) {
                    VirtualNetworkInfo network = new VirtualNetworkInfo();

                    VirtualTaskInfo taskInfo = new VirtualTaskInfo(dto.getTaskName());

                    String vmTemplate = "";
                    BufferedReader reader = null;
                    try {
                        reader = new BufferedReader(new FileReader("/media/Extra/SVN_GAMES_TRULLY_LATEST/GAMES-GlobalLoop/TUC_EPKB_Manager/vmTemplate.xml"));
                        String line = "";
                        while ((line = reader.readLine()) != null) {
                            vmTemplate += line + "\n";
                        }
                    } catch (Exception e) {
                        GlobalLoopLogger.getLogger().info(e.toString());
                        GlobalLoopLogger.error(e);
                    }

                    taskInfo.setVmTemplate(vmTemplate);
                    network.setIp(dto.getIp());
                    network.setVncPassword("123456");
                    network.setVncPort(dto.getVncPort());
                    taskInfo.setNetworkInfo(network);

                    for (ActivityDiskDto activityDiskDto : dto.getDiskDtoCollection()) {
                        VirtualDiskInfo virtualDiskInfo = new VirtualDiskInfo();
                        virtualDiskInfo.setFormat(activityDiskDto.getFormat());
                        virtualDiskInfo.setSize(activityDiskDto.getSize());
                        virtualDiskInfo.setSource(activityDiskDto.getPath());
                        virtualDiskInfo.setType(activityDiskDto.getType());
                        taskInfo.addDiskDto(virtualDiskInfo);
                    }

                    taskInfo.setRequestedCPU(dto.getRequestedCPUMax());
                    taskInfo.setRequestedMemory(dto.getRequestedMemoryMax());
                    try {
                        VirtualTaskInfo virtualTaskInfo = serviceCenterManagement.createVirtualMachine(taskInfo);
                        dto.setCloudID(virtualTaskInfo.getId());
                    } catch (ServiceCenterAccessException e) {
                        GlobalLoopLogger.getLogger().info(e.toString());
                        GlobalLoopLogger.error(e);
                    }
                }
            }

            //delete virtual machines
            for (ActivityDto dto : activityDtosToBeDeleted) {
                VirtualNetworkInfo network = new VirtualNetworkInfo();
                VirtualTaskInfo taskInfo = new VirtualTaskInfo(dto.getTaskName());
                taskInfo.setId(dto.getCloudID());
                network.setId(dto.getNetworkID());
                taskInfo.setNetworkInfo(network);
                try {
                    serviceCenterManagement.deleteVirtualMachine(taskInfo);
                } catch (ServiceCenterAccessException ex) {
                    Logger.getLogger(WorkloadSchedulerAgent.class.getName()).log(Level.SEVERE, null, ex);
                }
            }

        }

        if (!somethingWasCalled) {
            generateTasks();
        }

    }

//                generateTasks();
//            }
//        };
//        int generationTimerDelay = random.nextInt(WORKLOAD_GENERATION_INTERVAL_MAX - WORKLOAD_GENERATION_INTERVAL_MIN + 1) + WORKLOAD_GENERATION_INTERVAL_MIN;
//        System.out.println("Delaaaaaaaaaaay" + generationTimerDelay);
//        timer.schedule(task, generationTimerDelay * 1000);
    public void setPauseGenerating(boolean pauseGenerating) {
        this.pauseGenerating = pauseGenerating;
        if (!pauseGenerating) {
            generateTasks();
        }
    }

    private void addActivitiesAndPersistSchedule(List<ActivityDto> activities, long delay) {

        for (ActivityDto activityDto : activities) {

            ApplicationDescription applicationDescription = new ApplicationDescription();
            applicationDescription.setName(activityDto.getTaskName() + "Application");
            ActivityDescription activityDescription = new ActivityDescription();
            activityDescription.setName(activityDto.getTaskName() + "Activity");
            ActivityService activityService = new ActivityService();

            activityService.setServiceName(activityDto.getServiceName());
            activityService.setCpuWeight(activityDto.getCpuWeight());
            activityService.setIp(activityDto.getIp());
            activityService.setMemWeight(activityDto.getMemWeight());
            activityService.setRequestedCoresCount((float) activityDto.getRequestedCores());
            activityService.setRequestedCPUMax(activityDto.getRequestedCPUMax());
            activityService.setRequestedCPUMin(activityDto.getRequestedCPUMin());
            activityService.setRequestedMemoryMax(activityDto.getRequestedMemoryMax());
            activityService.setRequestedMemoryMin(activityDto.getRequestedMemoryMin());
            activityService.setRequestedStorageMax(activityDto.getRequestedStorageMax());
            activityService.setRequestedStorageMin(activityDto.getRequestedStorageMin());
            activityService.setStorageWeight(activityDto.getCpuWeight());
            activityService.setVncPort(activityDto.getVncPort());
            activityDescription.setDuration(activityDto.getLife());
            activityDescription.setServices(new ActivityService[]{activityService});

            ArrayList<ActivityDiskDto> activityDiskDtos = (ArrayList<ActivityDiskDto>) activityDto.getDiskDtoCollection();

            int disksCount = activityDiskDtos.size();
            VirtualDiskDescription[] virtualDisks = new VirtualDiskDescription[disksCount];

            for (int i = 0; i < disksCount; i++) {
                ActivityDiskDto activityDiskDto = activityDiskDtos.get(i);
                VirtualDiskDescription virtualDiskDescription = new VirtualDiskDescription();
                virtualDiskDescription.setFormat(activityDiskDto.getFormat());
                virtualDiskDescription.setSize(activityDiskDto.getSize());
                virtualDiskDescription.setTarget(activityDiskDto.getFormat());
                virtualDiskDescription.setType(activityDiskDto.getType());
                virtualDiskDescription.setSource(activityDiskDto.getPath());
                virtualDisks[i] = virtualDiskDescription;
            }

            activityService.setDisks(virtualDisks);
            applicationDescription.setActivities(new ActivityDescription[]{activityDescription});

            JaxbPair[] oldPairs = workloadSchedule.getApplications();
            int oldPairsSize = oldPairs.length;
            JaxbPair[] newPairs = new JaxbPair[oldPairsSize + 1];
            System.arraycopy(oldPairs, 0, newPairs, 0, oldPairsSize);


            newPairs[oldPairsSize] = new JaxbPair(delay, applicationDescription);

            workloadSchedule.setApplications(newPairs);
        }

        try {
            JAXBContext context = JAXBContext.newInstance(WorkloadSchedule.class);
            Marshaller marshaller = context.createMarshaller();
            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
            marshaller.marshal(workloadSchedule, generatedSchedule);
        } catch (Exception e) {
            e.printStackTrace();
        }

    }
}
