/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextAccessAndProcessing;

import java.util.Collection;
import java.util.List;

import ro.utcn.dsrl.contextAccessingAndProcessingAPI.utils.Pair;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.api.EACMInstance;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.GPI_KPI_Policy;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.ComplexResource;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.SimpleResource;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.util.exceptions.ContextModelAccessException;
import ro.utcn.dsrl.utils.commonDtos.ActivityDto;
import ro.utcn.dsrl.utils.commonDtos.ContextSituationDto;
import ro.utcn.dsrl.utils.exceptions.ApplicationException;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */

public interface ContextAccessAndProcessingAPI {
    //List<ActivityDto> getAllServices() throws  ContextModelAccessException;
    /**
     * @param eacmInstance the context model instance from which the workload is extracted
     * @param resource     a SimpleResource instance. The name of the SimpleResource is used for query
     * @return
     * @throws ContextModelAccessException if resource not found
     */
    Double getLoadForSimpleResource(EACMInstance eacmInstance, SimpleResource resource) throws ContextModelAccessException;

    /**
     * @param eacmInstance the context model instance from which the workload is extracted
     * @param resourceName the name of the SimpleResource for which the workload is retrieved
     * @return
     * @throws ContextModelAccessException if resource not found
     */
    Double getLoadForSimpleResource(EACMInstance eacmInstance, String resourceName) throws ContextModelAccessException;

    /**
     * @param eacmInstance    the context model instance from which the workload is extracted
     * @param complexResource a ComplexResource instance. The name of the ComplexResource is used for query
     * @return
     * @throws ContextModelAccessException if resource not found
     */
    List<Pair<SimpleResource, Double>> getLoadForComplexResource(EACMInstance eacmInstance, ComplexResource complexResource)
            throws ContextModelAccessException;

    /**
     * @param eacmInstance        eacmInstance the context model instance from which the workload is extracted
     * @param complexResourceName The name of the ComplexResource for which the workload is retrieved
     * @return
     * @throws ContextModelAccessException if resource not found
     */
    List<Pair<SimpleResource, Double>> getLoadForComplexResource(EACMInstance eacmInstance, String complexResourceName)
            throws ContextModelAccessException;

    /**
     * Compares the total number of running servers against the optimum number
     * specified by the Number
     *
     * @param eacmInstance the context model instance for which the metric is evaluated
     * @param threshold    a Number(Integer) value which represents the optimum number
     *                     of servers that should be running for the given EACMInstance workload
     * @return true if the number of running servers is <= threshold
     * @throws ContextModelAccessException currently it does not throw ContextModelAccessException.
     *                                     Placed here to accomodate more complex DH_UE evaluation in the future
     */
    Boolean evaluateMetricDH_UE(EACMInstance eacmInstance, Number threshold) throws ContextModelAccessException;

    /**
     * Computes the ratio between the number of online servers and the number of online servers which
     * perform useful activities( run virtual machines). The ratio is then compared with the
     * threshold value.
     *
     * @param eacmInstance the context model instance for which the metric is evaluated
     * @param threshold    a Number(Real) value which represents the optimum ratio between
     *                     the number of online servers and the number of online servers which run virtual
     *                     machines
     * @return true if the ratio between the online servers and the number of online servers which
     *         perform useful activities <= threshold and false otherwise
     * @throws ContextModelAccessException in case of SWRL policy evaluation failure.
     *                                     Example: If the EACMInstance provided is corrupted or does not contain all
     *                                     individuals(i.e. a older version)
     */
    Boolean evaluateMetricDH_UR(EACMInstance eacmInstance, Number threshold) throws ContextModelAccessException;

    /**
     * @param instance the context model instance on which the policy value is set
     * @param policy   the GPI_KPI policy for which the value is set. The <b>name</b>
     *                 is used for searching the GPI_KPI policy in the provided EACM instance
     * @param value    the GPI_KPI value
     * @throws ContextModelAccessException if policy not found
     */
    void setGPI_KPI_Value(EACMInstance instance, GPI_KPI_Policy policy, Number value) throws ContextModelAccessException;

    /**
     * @param instance   the context model instance on which the policy value is set
     * @param policyName used for searching the GPI_KPI policy in the provided EACM instance
     * @param value      the GPI_KPI value
     * @throws ContextModelAccessException if policy not found
     */
    void setGPI_KPI_Value(EACMInstance instance, String policyName, Number value) throws ContextModelAccessException;

    /**
     * @param instance          the context model instance on which the policy value is set
     * @param policy            the GPI_KPI policy for which the value is set. The <b>name</b>
     *                          is used for searching the GPI_KPI policy in the provided EACM instance
     * @param thresholdMinValue the minimum accepted value for the GPI_KPI
     * @param thresholdMaxValue the maximum accepted value for the GPI_KPI
     * @throws ContextModelAccessException if policy not found
     */
    void setGPI_KPI_Threshold(EACMInstance instance, GPI_KPI_Policy policy,
                              Number thresholdMinValue, Number thresholdMaxValue) throws ContextModelAccessException;

    /**
     * @param instance          the context model instance on which the policy value is set
     * @param policyName        used for searching the GPI_KPI policy in the provided EACM instance
     * @param thresholdMinValue the minimum accepted value for the GPI_KPI
     * @param thresholdMaxValue the maximum accepted value for the GPI_KPI
     * @throws ContextModelAccessException if policy not found
     */
    void setGPI_KPI_Threshold(EACMInstance instance, String policyName,
                              Number thresholdMinValue, Number thresholdMaxValue) throws ContextModelAccessException;

    /**
     * @return the current EACMInstance in use or null if no current
     *         EACMInstance was set(by performing a loadAndSetCurrent)
     */
    EACMInstance getCurrentContextInstance();

    /**
     * Creates an OntologyEACMInstance from a OWL model file
     *
     * @return an EACMInstance object created from the ontology file
     * @throws ContextModelAccessException
     */
    EACMInstance createContextInstance() throws ContextModelAccessException;

    /**
     * Creates an OntologyEACMInstance from a OWL model file
     *
     * @param owlFilePath Path to the OWL file containing the ontology
     * @return an EACMInstance object created from the ontology file
     * @throws ContextModelAccessException
     */
    EACMInstance createContextInstance(String owlFilePath) throws ContextModelAccessException;


    /**
     * Creates a local EACMInstance from a OWL file.
     *
     * @return an EACMInstance object created from the ontology file
     * @throws ContextModelAccessException
     */
    EACMInstance loadContextInstance(String owlFilePath) throws ContextModelAccessException;


    /**
     * Saves an OWL snapshot for the specified EACM instance to the specified path.
     *
     * @param path     the path towards the desired location of the .owl file to store the ontology
     * @param instance the ontology instance to be persisted
     * @throws ContextModelAccessException
     */
    void saveContextInstance(String path, EACMInstance instance) throws ContextModelAccessException;

    /**
     * Saves an OWL snapshot for the current EACMInstance to the specified directory
     *
     * @param saveDirectoryPath path to the directory which will contain all the EACMInstance snapshots
     * @throws ContextModelAccessException
     */
    void saveContextInstance(String saveDirectoryPath) throws ContextModelAccessException, ApplicationException;

    /**
     * @param dto information needed to uniquely identify the service in the service pool
     * @return information about a service from the service pool
     */
    public ActivityDto getService(ActivityDto dto) throws ContextModelAccessException;

    /**
     * @param dto the new service to be added to  the service pool
     */
    public void addService(ActivityDto dto) throws ContextModelAccessException;

    /**
     * @param dto information needed to uniquely identify the service in the service pool
     *            Removes a service from the service pool
     */
    public void removeService(ActivityDto dto) throws ContextModelAccessException;

    /**
     * @param dto information needed to uniquely identify a service in the deployed services pool
     * @return information about a deployed service ( a service running on the testbed )
     */
    public ActivityDto getDeployedService(ActivityDto dto) throws ContextModelAccessException;

    /**
     * @param dto information about a service running on the testbed which will be added in the deployed
     *            services list
     */
    public void addDeployedService(ActivityDto dto) throws ContextModelAccessException;

    /**
     * @param dto
     */
    public void removeDeployedService(ActivityDto dto) throws ContextModelAccessException;


    public void updateContextModelWithMonitoredData() throws ContextModelAccessException;

    /**
     *
     * @return
     * @throws ContextModelAccessException
     */
    public Collection<ContextSituationDto> getHistoricalContextSituations() throws ContextModelAccessException;

    public void persistHistoricalContextSituations(Collection<ContextSituationDto> contextSituationDtos) throws ContextModelAccessException;
    
    public void persistHistoricalContextSituation(ContextSituationDto contextSituationDto) throws ContextModelAccessException;

}
