/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.impl;

import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.*;
import edu.stanford.smi.protege.model.FrameID;
import edu.stanford.smi.protegex.owl.model.OWLModel;
import edu.stanford.smi.protegex.owl.model.RDFProperty;

import java.util.Collection;
import java.util.List;


/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */

public class DefaultServiceCenterServer extends DefaultComplexResource
        implements ServiceCenterServer {

    public DefaultServiceCenterServer(OWLModel owlModel, FrameID id) {
        super(owlModel, id);
    }


    public DefaultServiceCenterServer() {
    }

    public void markInitialValues() {
        if (isCluster()) {
            for (ServiceCenterITComputingResource resource : getResources()) {
                ((ServiceCenterServer) resource).markInitialValues();
            }
        } else {
            setOptimumCPU(((CPU) getCpuResources().iterator().next()).getOptimalWorkLoad().floatValue());
            setOptimumMEM(((MEM) getMemResources().iterator().next()).getOptimalWorkLoad().floatValue());
        }
    }


    public void resetInitialValues() {
        if (isCluster()) {
            for (ServiceCenterITComputingResource resource : getResources()) {
                ((ServiceCenterServer) resource).resetInitialValues();
            }
        } else {
            if (!hasOptimumCPU()) {
                markInitialValues();
            }
            float optimumCPU = getOptimumCPU();
            float optimumMEM = getOptimumMEM();
            CPU cpu = (CPU) this.getCpuResources().iterator().next();
            for (Core core : cpu.getAssociatedCores()) {
                core.setOptimalWorkLoad((double) optimumCPU);
            }
            cpu.setOptimalWorkLoad((double) optimumCPU);
            MEM mem = (MEM) this.getMemResources().iterator().next();
            mem.setOptimalWorkLoad((double) optimumMEM);
        }
    }

    public boolean hostsActivity(ApplicationActivity activity) {
        return getRunningActivities().contains(activity);
    }


    // Property http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#optimumCPU

    public float getOptimumCPU() {
        return getPropertyValueLiteral(getOptimumCPUProperty()).getFloat();
    }


    public RDFProperty getOptimumCPUProperty() {
        final String uri = "http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#optimumCPU";
        final String name = getOWLModel().getResourceNameForURI(uri);
        return getOWLModel().getRDFProperty(name);
    }


    public boolean hasOptimumCPU() {
        return getPropertyValueCount(getOptimumCPUProperty()) > 0;
    }


    public void setOptimumCPU(float newOptimumCPU) {
        setPropertyValue(getOptimumCPUProperty(), new java.lang.Float(newOptimumCPU));
    }


    // Property http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#optimumMEM

    public float getOptimumMEM() {
        return getPropertyValueLiteral(getOptimumMEMProperty()).getFloat();
    }


    public RDFProperty getOptimumMEMProperty() {
        final String uri = "http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#optimumMEM";
        final String name = getOWLModel().getResourceNameForURI(uri);
        return getOWLModel().getRDFProperty(name);
    }


    public boolean hasOptimumMEM() {
        return getPropertyValueCount(getOptimumMEMProperty()) > 0;
    }


    public void setOptimumMEM(float newOptimumMEM) {
        setPropertyValue(getOptimumMEMProperty(), new java.lang.Float(newOptimumMEM));
    }


    public void addRunningActivity(ApplicationActivity activity) {
//        if(this.getRunningActivities().contains(activity)){
//            return;
//        }
//        System.err.println(Thread.currentThread().getName());
        super.addRunningActivity(activity);
        activity.addAssociatedServer(this);

        if (isCluster()) {
            return;
        }
        Collection<CPU> cpus = this.getCpuResources();
        Collection<MEM> mems = this.getMemResources();
        Collection<HDD> hdds = this.getHddResources();

        for (CPU cpu : cpus) {
            if (cpu.hasResourcesFor(activity)) {
                cpu.addRunningActivity(activity);
                break;
            }
        }

        for (MEM mem : mems) {
            if (mem.hasResourcesFor(activity)) {
                mem.addRunningActivity(activity);
                break;
            }
        }

        for (HDD hdd : hdds) {
            if (hdd.hasResourcesFor(activity)) {
                hdd.addRunningActivity(activity);
            }
        }

        activity.setNumberOfCoresAllocatedValue(activity.getNumberOfCoresRequiredValue());
        activity.setCpuAllocatedValue(activity.getCpuRequiredMaxValue());
        activity.setMemAllocatedValue(activity.getMemRequiredMaxValue());
        activity.setHddAllocatedValue(activity.getHddRequiredMaxValue());

    }


    public void removeRunningActivity(ApplicationActivity activity) {

        super.removeRunningActivity(activity);
        activity.removeAssociatedServer(this);

        if (isCluster()) {
            for (ServiceCenterITComputingResource resource : getResources()) {
                if (resource.getRunningActivities().contains(activity)) {
                    resource.removeRunningActivity(activity);
                    break;
                }
            }
            return;
        }

        Collection<CPU> cpus = this.getCpuResources();
        Collection<MEM> mems = this.getMemResources();
        Collection<HDD> hdds = this.getHddResources();

        for (CPU cpu : cpus) {
            if (cpu.getRunningActivities().contains(activity)) {
                cpu.removeRunningActivity(activity);
                activity.removeAllReceivedCoreIndex();
                break;
            }
        }

        for (MEM mem : mems) {
            if (mem.getRunningActivities().contains(activity)) {
                mem.removeRunningActivity(activity);
                break;
            }
        }

        for (HDD hdd : hdds) {
            if (hdd.getRunningActivities().contains(activity)) {
                hdd.removeRunningActivity(activity);
            }
        }


        activity.setCpuAllocatedValue(0);
        activity.setMemAllocatedValue(0);
        activity.setHddAllocatedValue(0);


    }


    public String toString() {
        String description = "";
        description += this.getLocalName() + "\n";
        description += "Active:" + this.getIsActive() + "\n";
        if (isCluster()) {
            description += " [ ";
            Collection<ServiceCenterITComputingResource> resources = this.getResources();
            for (ServiceCenterITComputingResource resource : resources) {
                description += ((ServiceCenterServer) resource).toString();
            }
            description += " ] ";
            return description;
        } else {
            Collection<ServiceCenterITComputingResource> resources = this.getMemResources();
            for (ServiceCenterITComputingResource resource : resources) {
                description += " " + resource.getLocalName();
                description += " maxWorkload: " + resource.getMaximumWorkLoad();
                description += " currentWorkload: " + resource.getCurrentWorkLoad() + "\n";
            }

            resources = this.getHddResources();
            for (ServiceCenterITComputingResource resource : resources) {
                description += " " + resource.getLocalName();
                description += " maxWorkload: " + resource.getMaximumWorkLoad();
                description += " currentWorkload: " + resource.getCurrentWorkLoad() + "\n";
            }

            List<Core> cores = ((CPU) this.getCpuResources().iterator().next()).getAssociatedCores();
            for (Core resource : cores) {
                description += " " + resource.getLocalName();
                description += " maxWorkload: " + resource.getMaximumWorkLoad();
                description += " currentWorkload: " + resource.getCurrentWorkLoad() + "\n";
            }
        }
        return description;
    }

    // Property http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#ipAddress

    public String getIpAddress() {
        return (String) getPropertyValue(getIpAddressProperty());
    }


    public RDFProperty getIpAddressProperty() {
        final String uri = "http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#ipAddress";
        final String name = getOWLModel().getResourceNameForURI(uri);
        return getOWLModel().getRDFProperty(name);
    }


    public boolean hasIpAddress() {
        return getPropertyValueCount(getIpAddressProperty()) > 0;
    }


    public void setIpAddress(String newIpAddress) {
        setPropertyValue(getIpAddressProperty(), newIpAddress);
    }


    // Property http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#macAddress

    public String getMacAddress() {
        return (String) getPropertyValue(getMacAddressProperty());
    }


    public RDFProperty getMacAddressProperty() {
        final String uri = "http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#macAddress";
        final String name = getOWLModel().getResourceNameForURI(uri);
        return getOWLModel().getRDFProperty(name);
    }


    public boolean hasMacAddress() {
        return getPropertyValueCount(getMacAddressProperty()) > 0;
    }


    public void setMacAddress(String newMacAddress) {
        setPropertyValue(getMacAddressProperty(), newMacAddress);
    }

//    http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#serverPowerConsumptionWhenIdle

    public RDFProperty getPowerConsumptionWhileIdleProperty() {
        final String uri = "http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#serverPowerConsumptionWhenIdle";
        final String name = getOWLModel().getResourceNameForURI(uri);
        return getOWLModel().getRDFProperty(name);
    }

    public RDFProperty getPowerConsumptionWhileLoadedTo40Property() {
        final String uri = "http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#serverPowerConsumptionWhenLoadedTo40";
        final String name = getOWLModel().getResourceNameForURI(uri);
        return getOWLModel().getRDFProperty(name);
    }

    public RDFProperty getPowerConsumptionWhileLoadedTo100Property() {
        final String uri = "http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#serverPowerConsumptionWhenLoadedTo100";
        final String name = getOWLModel().getResourceNameForURI(uri);
        return getOWLModel().getRDFProperty(name);
    }

    public int getPowerConsumptionWhileIdle() {
        return (Integer) getPropertyValue(getPowerConsumptionWhileIdleProperty());
    }

    public int getPowerConsumptionWhileLoadedTo40() {
        return (Integer) getPropertyValue(getPowerConsumptionWhileLoadedTo40Property());
    }

    public int getPowerConsumptionWhileLoadedTo100() {
        return (Integer) getPropertyValue(getPowerConsumptionWhileLoadedTo100Property());
    }

    public void setPowerConsumptionWhileIdle(int value) {
        setPropertyValue(getPowerConsumptionWhileIdleProperty(), value);
    }

    public void setPowerConsumptionWhileLoadedTo40(int value) {
        setPropertyValue(getPowerConsumptionWhileLoadedTo40Property(), value);
    }

    public void setPowerConsumptionWhileLoadedTo100(int value) {
        setPropertyValue(getPowerConsumptionWhileLoadedTo100Property(), value);
    }

    // Property http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#serviceCenterServerPolicy

    public ITComputingContextPolicy getServiceCenterServerPolicy() {
        return (ITComputingContextPolicy) getPropertyValueAs(getServiceCenterServerPolicyProperty(), ITComputingContextPolicy.class);
    }


    private RDFProperty getServiceCenterServerPolicyProperty() {
        final String uri = "http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#serviceCenterServerPolicy";
        final String name = getOWLModel().getResourceNameForURI(uri);
        return getOWLModel().getRDFProperty(name);
    }


    public boolean hasServiceCenterServerPolicy() {
        return getPropertyValueCount(getServiceCenterServerPolicyProperty()) > 0;
    }


    public void setServiceCenterServerPolicy(ITComputingContextPolicy newServiceCenterServerPolicy) {
        setPropertyValue(getServiceCenterServerPolicyProperty(), newServiceCenterServerPolicy);
    }


}
