/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.impl;

import java.util.Collection;

import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.*;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.ServerLoadPolicyViolationEvaluator;
import ro.utcn.dsrl.utils.config.GeneralConfigurationManager;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */


public class DefaultServerLoadPolicyViolationEvaluator implements ServerLoadPolicyViolationEvaluator {

    public double evaluateServerLoadPolicyViolationDegree(ServiceCenterServer server) {
        double respectance = 0.0;

        CPU associatedCPU = null;
        HDD storage = null;
        MEM memory = null;
        double diff = 0;
        if (server.getIsActive()) {
            for (CPU cpu : server.getCpuResources()) {
                Double usedCPU = cpu.getCurrentWorkLoad();
                Double totalCPU = cpu.getMaximumWorkLoad();
                Double optimalCPU = cpu.getOptimalWorkLoad();
                
                Double sumFirstElement = totalCPU*GeneralConfigurationManager.getResourceLoadOptimalValue()/100 ;
                Double sumSecondElement = GeneralConfigurationManager.getResourceLoadVariation() * totalCPU / 100 ;
                
                Double cpuMaxAcceptableValue = sumFirstElement + sumSecondElement ;
                Double cpuMinAcceptableValue = sumFirstElement - sumSecondElement ;
                
                if (usedCPU < cpuMinAcceptableValue) {
                    diff += cpuMinAcceptableValue - usedCPU;
                }
                if (usedCPU > cpuMaxAcceptableValue) {
                    diff += usedCPU - cpuMaxAcceptableValue;
                }
                if (server.hasCPUWeight()) {
                    respectance += server.getCPUWeight() * diff;
                } else {
                    respectance += diff;
                }
            }

            diff = 0.0;
            double usedMemory = 0.0;
            double total = 0.0;
            double optimal = 0.0;
            double memoryMaxAcceptableValue = 0.0;

            //when deployed as scheduler it is difficult to read the CPU requirements
            if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                double memoryMinAcceptableValue = optimal / 2.0;
                Collection<MEM> memories = server.getMemResources();
                for (MEM mem : memories) {
                    usedMemory = mem.getCurrentWorkLoad();
                    total = mem.getMaximumWorkLoad();
                    optimal = mem.getOptimalWorkLoad();
                    memoryMaxAcceptableValue = total*GeneralConfigurationManager.getResourceLoadOptimalValue()/100 + GeneralConfigurationManager.getResourceLoadVariation() * total / 100;
                    memoryMinAcceptableValue =  total*GeneralConfigurationManager.getResourceLoadOptimalValue()/100 - GeneralConfigurationManager.getResourceLoadVariation() * total / 100;
                    if (usedMemory > memoryMaxAcceptableValue) {
                        diff = usedMemory - memoryMaxAcceptableValue;
                    } else if (usedMemory < memoryMinAcceptableValue) {
                        diff = memoryMinAcceptableValue - usedMemory;
                    }
                    if (server.hasMEMWeight()) {
                        respectance += server.getMEMWeight() * diff;
                    } else {
                        respectance += diff;
                    }
                }
                diff = 0.0;

                Double usedStorage = 0.0;
                Double storageMaxAcceptableValue = 0.0;
                double storageMinAcceptableValue = 0.0;

                for (HDD hdd : server.getHddResources()) {
                    total = hdd.getMaximumWorkLoad();
                    optimal = hdd.getOptimalWorkLoad();
                    usedStorage = hdd.getCurrentWorkLoad();
                    storageMaxAcceptableValue =  total*GeneralConfigurationManager.getResourceLoadOptimalValue()/100 + GeneralConfigurationManager.getResourceLoadVariation() * total / 100;
                    ;
                    storageMinAcceptableValue =  total*GeneralConfigurationManager.getResourceLoadOptimalValue()/100 - GeneralConfigurationManager.getResourceLoadVariation() * total / 100;
                    ;
                    if (usedStorage > storageMaxAcceptableValue) {
                        diff = usedStorage - storageMaxAcceptableValue;
                    } else if (usedStorage < storageMinAcceptableValue) {
                        diff = storageMinAcceptableValue - usedStorage;
                    }
                    if (server.hasHDDWeight()) {
                        respectance += server.getHDDWeight() * diff;
                    } else {
                        respectance += diff;
                    }
                }
            }
        }
        return respectance;
    }

    public double getMinWorkload(SimpleResource simpleResource) {
        Double total = simpleResource.getMaximumWorkLoad();
        Double optimal = simpleResource.getOptimalWorkLoad();
        Double minAcceptableValue = optimal - GeneralConfigurationManager.getResourceLoadVariation() * total / 100;

        return minAcceptableValue;
    }

    public double getMaxWorkload(SimpleResource simpleResource) {
        Double total = simpleResource.getMaximumWorkLoad();
        Double optimal = simpleResource.getOptimalWorkLoad();
        Double maxAcceptableValue = optimal + GeneralConfigurationManager.getResourceLoadVariation() * total / 100;
        return maxAcceptableValue;
    }
}
