/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.impl;


import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.api.EACMInstance;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.CPU;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.ServiceCenterServer;
import ro.utcn.dsrl.utils.commonInterfaces.PowerConsumption;
import ro.utcn.dsrl.utils.exceptions.ApplicationException;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Collection;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */


public class PowerConsumptionEstimator extends UnicastRemoteObject implements PowerConsumption {
    private EACMInstance modelAccess;
    public static final int BASE_POWER_CONSUMPTION = 70;
    public static final int POWER_CONSUMPTION_TO_40 = 100;
    public static final int POWER_CONSUMPTION_TO_100 = 130;
    public static final int POWER_CONSUMPTION_IN_HIBERNATE = 4;
    public static final int GLOBAL_LOOP_OVERHEAD = 30;
    public static final int BASE_CONTROLLER_CONSUMPTION = 110;
    public static final int POWER_CONSUMPTION_ON_DEPLOY = 10;
    public static final int POWER_CONSUMPTION_ON_MIGRATE = 20;
    public boolean isSimulation = false;
    private double previousPowerConsumptionWithRTE = 0;
    private double previousPowerConsumptionWithoutRTE = 0;
    private int numberOfDeployActions = 0;
    private int numberOfMigrateActions = 0;

    public PowerConsumptionEstimator(EACMInstance modelAccess) throws RemoteException {
        this.modelAccess = modelAccess;
    }

    public Number getValueWithRunningAlgorithm() throws ApplicationException {
        double toReturn = 0.0;

        if (isSimulation) {

            return previousPowerConsumptionWithRTE;

        } else {

            toReturn += POWER_CONSUMPTION_ON_DEPLOY * numberOfDeployActions;
            toReturn += POWER_CONSUMPTION_ON_MIGRATE * numberOfMigrateActions;

            Collection<ServiceCenterServer> servers = modelAccess.getAllServiceCenterServerInstances();

            for (ServiceCenterServer server : servers) {
                if (server.getIsActive()) {
                    CPU cpu = server.getCpuResources().iterator().next();
                    //TODO: de introdus in contextModel si proprietatile de energyConsumption
                    if (cpu.getCurrentWorkLoad() * 100 / cpu.getMaximumWorkLoad() < 20)
                        toReturn += BASE_POWER_CONSUMPTION;
                    else if (cpu.getCurrentWorkLoad() * 100 / cpu.getMaximumWorkLoad() < 40)
                        toReturn += POWER_CONSUMPTION_TO_40;
                    else
                        toReturn += POWER_CONSUMPTION_TO_100;
                } else {
                    toReturn += POWER_CONSUMPTION_IN_HIBERNATE;
                }
            }
            toReturn += GLOBAL_LOOP_OVERHEAD; //Global Loop RTE overhead
            toReturn += BASE_CONTROLLER_CONSUMPTION;
            previousPowerConsumptionWithRTE = toReturn;
        }
        return toReturn;
    }

    public Number getValueWithoutAlgorithm() throws ApplicationException {
        double toReturn = 0.0;
        if (isSimulation) {
            return previousPowerConsumptionWithoutRTE;
        } else {
            toReturn += POWER_CONSUMPTION_ON_DEPLOY * numberOfDeployActions;
            Collection<ServiceCenterServer> servers = modelAccess.getAllServiceCenterServerInstances();
            for (ServiceCenterServer server : servers) {
                if (server.getIsActive()) {
                    CPU cpu = server.getCpuResources().iterator().next();
                    if (cpu.getCurrentWorkLoad() * 100 / cpu.getMaximumWorkLoad() < 20) {
                        toReturn += BASE_POWER_CONSUMPTION;
                    } else if (cpu.getCurrentWorkLoad() * 100 / cpu.getMaximumWorkLoad() < 40) {
                        toReturn += POWER_CONSUMPTION_TO_40;
                    } else {
                        toReturn += POWER_CONSUMPTION_TO_100;
                        ;
                    }
                } else {
                    toReturn += BASE_POWER_CONSUMPTION;
                }
            }
            toReturn += BASE_CONTROLLER_CONSUMPTION;
            previousPowerConsumptionWithoutRTE = toReturn;
        }
        return toReturn;
    }

    public boolean isSimulation() throws ApplicationException {
        return isSimulation;
    }

    public void setSimulation(boolean t) throws ApplicationException {
        isSimulation = t;
    }

    public void setNumberOfDeployActions(int nr) throws RemoteException {
        numberOfDeployActions = nr;
    }


    public int getNumberOfDeployActions() throws RemoteException {
        return numberOfDeployActions;
    }

    public int getNumberOfMigrateActions() throws RemoteException {
        return numberOfMigrateActions;
    }

    public void setNumberOfMigrateActions(int numberOfMigrateActions) throws RemoteException {
        this.numberOfMigrateActions = numberOfMigrateActions;
    }
}
