/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
 
package ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.impl;

import edu.stanford.smi.protegex.owl.model.OWLModel;
import edu.stanford.smi.protegex.owl.swrl.exceptions.SWRLFactoryException;
import edu.stanford.smi.protegex.owl.swrl.model.SWRLFactory;
import edu.stanford.smi.protegex.owl.swrl.model.SWRLImp;
import edu.stanford.smi.protegex.owl.swrl.parser.SWRLParseException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.ApplicationActivity;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.ContextAction;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.ContextResource;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.GPI_KPI_Policy;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.ITComputingContextPolicy;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.QoSPolicy;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.ServiceCenterServer;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.util.exceptions.IndividualNotFoundException;
import ro.utcn.dsrl.contextAccessingAndProcessingAPI.utils.PolicyEvaluator;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.ContextEntropyEvaluator;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.QoSPolicyViolationEvaluator;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.contextMetricsComputation.ServerLoadPolicyViolationEvaluator;
import ro.utcn.dsrl.contextMonitoringModule.utils.objectToDTOMappers.RLNode;
import ro.utcn.dsrl.utils.config.GeneralConfigurationManager;
import ro.utcn.dsrl.utils.misc.Pair;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */
public class SWRLContextEntropyEvaluator implements ContextEntropyEvaluator {

    private ServerLoadPolicyViolationEvaluator serverLoadViolationEvaluator;
    private QoSPolicyViolationEvaluator qosPolicyViolationEvaluator;
    private static final String IS_RESPECTED_PROPERTY = "http://www.semanticweb.org/ontologies/2010/6/ContextModel.owl#isRespected";

    public SWRLContextEntropyEvaluator(ServerLoadPolicyViolationEvaluator serverLoadViolationEvaluator,
            QoSPolicyViolationEvaluator qosPolicyViolationEvaluator) {
        super();

        this.serverLoadViolationEvaluator = serverLoadViolationEvaluator;
        this.qosPolicyViolationEvaluator = qosPolicyViolationEvaluator;
    }

    public Pair<Double, List<GPI_KPI_Policy>> computeEntropy(
            Collection<GPI_KPI_Policy> policies) {
        List<GPI_KPI_Policy> brokenPolicies = new ArrayList<GPI_KPI_Policy>();
        GPI_KPI_Policy brokenPolicy = null;
        double entropy = 0.0;
        if (policies == null || policies.isEmpty()) {
            return new Pair<Double, List<GPI_KPI_Policy>>(entropy, brokenPolicies);
        }

        OWLModel model = policies.iterator().next().getOWLModel();
        
        SWRLFactory factory = new SWRLFactory(model);


        //just to make sure, delete any existing SWRL rules
//        for (SWRLImp imp : factory.getImps()) {
//            try {
//                factory.deleteImp(imp.getName());
//            } catch (SWRLFactoryException ex) {
//                Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
//            }
//        }


        //create SWRL for optimum workload for Server and QoS for applicatiopn activoty

        //if server is Off, coresponding energy policy is respected
        SWRLImp serverOffRule = null;
        try {
            //avoid creating the rules every time
            if (factory.hasImp("InactiveServersResourceUsageSWRLRule")) {
                serverOffRule = factory.getImp("InactiveServersResourceUsageSWRLRule");
            } else {
                serverOffRule = factory.createImp("InactiveServersResourceUsageSWRLRule", "ITComputingContextPolicy(?policy) "
                        + " ^ policySubject(?policy,?x) "
                        + " ^ ServiceCenterServer(?x) "
                        + " ^ isActive(?x,?isAlive) "
                        + " ^ swrlb:equal(?isAlive,false)"
                        + " -> "
                        + "isRespected(?policy,true)");
            }
            serverOffRule.enable();
        } catch (SWRLParseException ex) {
            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
        } catch (SWRLFactoryException ex) {
            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
        }

        float acceptedDeviation = GeneralConfigurationManager.getResourceLoadVariation();

        //if server is On, optimum workload indicators must be checked
        SWRLImp serverOnRule = null;
        try {
            if (factory.hasImp("ActiveServersResourceUsageSWRLRule")) {
                serverOnRule = factory.getImp("ActiveServersResourceUsageSWRLRule");
            } else {
                serverOnRule = factory.createImp("ActiveServersResourceUsageSWRLRule", "ITComputingContextPolicy(?policy) "
                        + " ^ policySubject(?policy,?x) "
                        + " ^ ServiceCenterServer(?x) "
                        + " ^ isActive(?x,?isAlive) "
                        + " ^ swrlb:equal(?isAlive,true)"
                        //get each CPU
                        + " ^ cpuResources(?x,?cpu)"
                        //check if Core max and optimal workload indicators are respected
                        + " ^ associatedCores(?cpu,?core)"
                        //check if Core max and optimal workload indicators are respected

                        //check if CPU workload is within limits

                        //get current, min and optimal workload values
                        + " ^ currentWorkload(?core,?currentCPUWorkloadVar)"
                        + " ^ maximumWorkload(?core,?maximumCPUWorkloadVar)"
                        + " ^ optimalWorkload(?core,?optimalCPUWorkloadVar)"
                        //compute percentage value of optimal workload and acepted deviation
                        + " ^ swrlb:divide(?optimalCPUWorkloadPercentage,?optimalCPUWorkloadVar,100.0)"
                        + " ^ swrlb:divide(?acceptedCPUDeviationPercentage," + acceptedDeviation + ",100.0)"
                        //                    
                        //compute minimum and maximum workload values considering the permited deviation
                        + " ^ swrlb:multiply(?firstCPUFunctionEl,?optimalCPUWorkloadPercentage,?maximumCPUWorkloadVar)"
                        + " ^ swrlb:multiply(?secondCPUFunctionEl,?acceptedCPUDeviationPercentage,?maximumCPUWorkloadVar)"
                        //                    
                        //compute max acceptable workload considering accepted deviation
                        + " ^ swrlb:add(?maximumCPUAcceptableWorkload,?firstCPUFunctionEl,?secondCPUFunctionEl)"
                        //compute min acceptable workload considering accepted deviation
                        + " ^ swrlb:subtract(?minimumCPUAcceptableWorkload,?firstCPUFunctionEl,?secondCPUFunctionEl)"
                        //                    compare current workload with max and min
                        + " ^ swrlb:greaterThan(?currentCPUWorkloadVar,?minimumCPUAcceptableWorkload)"
                        + " ^ swrlb:lessThan(?currentCPUWorkloadVar,?maximumCPUAcceptableWorkload)"
                        //get MEM resource
                        + " ^ memResources(?x,?memoryResource)"
                        //check if MEM workload is within limits

                        //get current, min and optimal workload values
                        + " ^ currentWorkload(?memoryResource,?currentMEMWorkloadVar)"
                        + " ^ maximumWorkload(?memoryResource,?maximumMEMWorkloadVar)"
                        + " ^ optimalWorkload(?memoryResource,?optimalMEMWorkloadVar)"
                        //compute percentage value of optimal workload and acepted deviation
                        + " ^ swrlb:divide(?optimalMEMWorkloadPercentage,?optimalMEMWorkloadVar,100.0)"
                        + " ^ swrlb:divide(?acceptedMEMDeviationPercentage," + acceptedDeviation + ",100.0)"
                        //                    
                        //compute minimum and maximum workload values considering the permited deviation
                        + " ^ swrlb:multiply(?firstMEMFunctionEl,?optimalMEMWorkloadPercentage,?maximumMEMWorkloadVar)"
                        + " ^ swrlb:multiply(?secondMEMFunctionEl,?acceptedMEMDeviationPercentage,?maximumMEMWorkloadVar)"
                        //                    
                        //compute max acceptable workload considering accepted deviation
                        + " ^ swrlb:add(?maximumMEMAcceptableWorkload,?firstMEMFunctionEl,?secondMEMFunctionEl)"
                        //compute min acceptable workload considering accepted deviation
                        + " ^ swrlb:subtract(?minimumMEMAcceptableWorkload,?firstMEMFunctionEl,?secondMEMFunctionEl)"
                        //                    compare current workload with max and min
                        + " ^ swrlb:greaterThan(?currentMEMWorkloadVar,?minimumMEMAcceptableWorkload)"
                        + " ^ swrlb:lessThan(?currentMEMWorkloadVar,?maximumMEMAcceptableWorkload)"
                        //set isRespected property for th policy to true
                        + " -> isRespected(?policy,true)");
            }
            serverOnRule.enable();
        } catch (SWRLParseException ex) {
            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
        } catch (SWRLFactoryException ex) {
            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
        }


        // Resource allocation (QoS) indicators must be checked for application activities
        SWRLImp applicationActivityRule = null;
        try {
            if (factory.hasImp("ActivitiesResourceUsageSWRLRule")) {
                applicationActivityRule = factory.getImp("ActivitiesResourceUsageSWRLRule");
            } else {
                applicationActivityRule = factory.createImp("ActivitiesResourceUsageSWRLRule", "QoSPolicy(?policy)"
                        + " ^ policySubject(?policy,?activityVar)"
                        + " ^ ApplicationActivity(?activityVar)"
                        //retrieve the cpuAllocated value (current cpu workload)
                        + " ^ cpuAllocatedValue(?activityVar,?allocatedCPU)"
                        //get maximum and minimum requested CPU
                        + " ^ cpuRequiredMaxValue(?activityVar,?maxRequiredCPU)"
                        + " ^ cpuRequiredMinValue(?activityVar,?minRequiredCPU)"
                        //retrieve the allocated memory value (current mem workload)
                        + " ^ memAllocatedValue(?activityVar,?allocatedMEM)"
                        //get maximum and minimum requested memory
                        + " ^ memRequiredMaxValue(?activityVar,?maxRequiredMEM)"
                        + " ^ memRequiredMinValue(?activityVar,?minRequiredMEM)"
                        //verify if allocatedCPU < maxRequiredCPU && allocatedCPU > minRequiredCPU
                        + " ^ swrlb:greaterThan(?allocatedCPU,?minRequiredCPU)"
                        + " ^ swrlb:lessThan(?allocatedCPU,?maxRequiredCPU)"
                        //verify if allocatedMEM < maxRequiredMEM && allocatedMEM > minRequiredMEM
                        + " ^ swrlb:greaterThan(?allocatedMEM,?minRequiredMEM)"
                        + " ^ swrlb:lessThan(?allocatedMEM,?maxRequiredMEM)"
                        //set isRespected property for th policy to true
                        + " -> isRespected(?policy,true)");
            }
            applicationActivityRule.enable();
        } catch (SWRLParseException ex) {
            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
        } catch (SWRLFactoryException ex) {
            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
        }


        //create new SWRL policy evaluator over the latest OntModel
        PolicyEvaluator policyAnalyser = new PolicyEvaluator(model.getOntModel());

        for (GPI_KPI_Policy policy : policies) {


            if (policy instanceof QoSPolicy) {
                Collection<ContextResource> task = policy.getPolicySubject();

                // if task has been deleted
                if (task == null) {
                    continue;
                }
                try {
                    if (!policyAnalyser.evaluateProperty(policy, IS_RESPECTED_PROPERTY)) {
                        brokenPolicies.add(policy);
                        if (brokenPolicy == null) {
                            brokenPolicy = policy;
                        }
                        if (policy.hasPolicyWeight()) {
                            for (ContextResource app : task) {
                                entropy += policy.getPolicyWeight()
                                        * qosPolicyViolationEvaluator.evaluateQoSPolicyViolationDegree((ApplicationActivity) app);
                            }
                        }
                    }
                } catch (IndividualNotFoundException ex) {
                    Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
                }

            } else if (policy instanceof ITComputingContextPolicy) {

                Collection<ContextResource> servers = policy.getPolicySubject();
                for (ContextResource r : servers) {
                    ServiceCenterServer server = (ServiceCenterServer) r;
                    if (server.getIsActive()) {
                        try {
                            if (!policyAnalyser.evaluateProperty(policy, IS_RESPECTED_PROPERTY)) {

                                brokenPolicies.add(policy);
                                if (brokenPolicy == null) {
                                    brokenPolicy = policy;
                                }
                                if (policy.hasPolicyWeight()) {
                                    entropy += policy.getPolicyWeight()
                                            * serverLoadViolationEvaluator.evaluateServerLoadPolicyViolationDegree(server);
                                }

                                //                            else {
                                //                                entropy += serverLoadViolationEvaluator.evaluateServerLoadPolicyViolationDegree((ServiceCenterServer) r);
                                //                            }else {
                                //                                entropy += serverLoadViolationEvaluator.evaluateServerLoadPolicyViolationDegree((ServiceCenterServer) r);
                                //                            }

                            }
                        } catch (IndividualNotFoundException ex) {
                            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
                        }

                    }
                }
            }
        }

//        //delete the created SWRL rules as not to interfere with or slow down the ontology interraction
//        //separate try-catch helps in the case that one is null (creation error) the others can be deleted
//        try {
//            factory.deleteImp(serverOffRule.getName());
//        } catch (SWRLFactoryException ex) {
//            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
//        }
//        try {
//            factory.deleteImp(serverOnRule.getName());
//        } catch (SWRLFactoryException ex) {
//            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
//        }
//        try {
//            factory.deleteImp(applicationActivityRule.getName());
//        } catch (SWRLFactoryException ex) {
//            Logger.getLogger(SWRLContextEntropyEvaluator.class.getName()).log(Level.SEVERE, null, ex);
//        }
        
        //disable imps 
        serverOffRule.disable();
        serverOnRule.disable();
        applicationActivityRule.disable();
        System.gc();

        return new Pair<Double, List<GPI_KPI_Policy>>(entropy, brokenPolicies);
    }

    public double computeRewardFunction(RLNode previous,
            RLNode current, List<ContextAction> contextActions) {
        double function;

        if (previous != null && previous.getActions().size() > 0) {
            function = previous.getRewardFunction();
        } else {
            function = 0;
        }
        if (previous != null) {
            int cost = 0;
            for (ContextAction action : contextActions) {
                cost += action.getCost();
            }
            double temp = previous.getContextEntropy()
                    - current.getContextEntropy() - cost; // - c.getCost() -
            // current.getActions().size()
            // * 100;
            function += RLNode.gamma * temp;

        } else {
            function = -current.getContextEntropy();
        }

        return function;
    }
}
