/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.impl;

import ro.utcn.dsrl.contextAccessingAndProcessingAPI.contextModel.concepts.*;
import ro.utcn.dsrl.contextAnalysisModule.contextAnalysis.PolicyAnalyser;
import ro.utcn.dsrl.utils.logging.GlobalLoopLogger;

import java.util.Collection;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */


public class DefaultPolicyAnalyser implements PolicyAnalyser {


    public boolean isPolicyRespected(ContextPolicy policy) {
        boolean respected = true;
        if (policy instanceof BusinessPolicy) {
            BusinessPolicy businessPolicy = (BusinessPolicy) policy;
            if (businessPolicy.getPolicySubject().size() == 0) {
                return true;
            }
            ApplicationActivity activity = (ApplicationActivity) businessPolicy.getPolicySubject().get(0);
            respected = (activity.getCpuRequiredMaxValue() >= activity.getCpuAllocatedValue()
                    && activity.getCpuRequiredMinValue() <= activity.getCpuAllocatedValue())
                    && (activity.getMemRequiredMaxValue() >= activity.getMemAllocatedValue()
                    && activity.getMemRequiredMinValue() <= activity.getMemAllocatedValue());
        } else if (policy instanceof ITComputingContextPolicy) {
            ITComputingContextPolicy contextPolicy = (ITComputingContextPolicy) policy;
            ServiceCenterServer server = (ServiceCenterServer) contextPolicy.getPolicySubject().get(0);

            if (server.isCluster()) {
                final AtomicBoolean atomicBoolean = new AtomicBoolean(true);
                Collection<ServiceCenterITComputingResource> resources = server.getResources();
                final CountDownLatch countDownLatch = new CountDownLatch(resources.size());

                for (final ServiceCenterITComputingResource resource : resources) {
                    Thread thread = new Thread("ThreadForEvaluatingPolicyFor " + resource.getName().split("#")[1]) {
                        @Override
                        public void run() {
                            if (!atomicBoolean.get()) {
                                countDownLatch.countDown();
                                return;
                            } else {
                                atomicBoolean.set(isPolicyRespected(((ServiceCenterServer) resource).getServiceCenterServerPolicy()));
                                countDownLatch.countDown();
                            }
                        }
                    };
                    thread.start();
                }
                try {
                    countDownLatch.await();
                } catch (InterruptedException e) {
                    GlobalLoopLogger.error(e);
                }
                return atomicBoolean.get();
            } else {
                if (!server.getIsActive()) {
                    return true;
                }
                CPU cpu = server.getCpuResources().iterator().next();
                MEM mem = server.getMemResources().iterator().next();

                if ((cpu.getCurrentWorkLoad() > (2 * cpu.getMaximumWorkLoad() + cpu.getOptimalWorkLoad()) / 3.0) || (cpu.getCurrentWorkLoad() < cpu.getOptimalWorkLoad() / 2.0))
                    return false;

                if ((mem.getCurrentWorkLoad() > mem.getMaximumWorkLoad() - (mem.getMaximumWorkLoad() - mem.getOptimalWorkLoad()) / 3.0)
                        || (mem.getCurrentWorkLoad() < mem.getOptimalWorkLoad() / 2.0)) {
                    return false;
                }
            }
        }
        return respected;
    }

}