/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.contextMonitoringModule.exposedFunctionality.rmiAPI.impl;

import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jade.wrapper.AgentController;
import ro.utcn.dsrl.mainModule.agents.MainAgent;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.ServiceCenterManagement;
import ro.utcn.dsrl.actionEnforcementModule.serviceCenterManagement.CAPInteractions;
import ro.utcn.dsrl.contextMonitoringModule.agents.ContextMonitoringAgent;
import ro.utcn.dsrl.utils.exposedFunctionality.jaxbBindingClasses.*;
import ro.utcn.dsrl.utils.exposedFunctionality.rmiAPI.RMIAPI;
import ro.utcn.dsrl.utils.exposedFunctionality.rmiAPI.RMICommand;
import ro.utcn.dsrl.utils.agentsCommunication.AgentMessageType;
import ro.utcn.dsrl.utils.agentsCommunication.InterAgentMessageBuilder;
import ro.utcn.dsrl.utils.commonDtos.*;
import ro.utcn.dsrl.utils.config.GeneralConfigurationManager;
import ro.utcn.dsrl.utils.exceptions.ApplicationException;
import ro.utcn.dsrl.utils.exceptions.ServiceCenterAccessException;
import ro.utcn.dsrl.utils.logging.GlobalLoopLogger;

import javax.xml.bind.JAXB;
import java.io.*;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

/**
 * <br/>
 * <br/>
 * <p>
 * <strong>Technical University of Cluj Napoca</strong>
 * </p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research
 * Laboratory </a> <br/>
 * Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro <br/>
 * <br/>
 * Developers: <br/>
 * Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro) <br/>
 * Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */
public class RMIInterface extends UnicastRemoteObject implements Serializable, RMIAPI {

    private jade.core.Agent agent;
    private ServiceCenterManagement serviceCenterManagement;
    private CAPInteractions capInteractions;
    private AgentController wsAgentController;
    private static int taskIndex = 0;

    public RMIInterface(Agent agent,
            ServiceCenterManagement serviceCenterManagement,
            CAPInteractions capInteractions)
            throws RemoteException {
        this.agent = agent;
        this.serviceCenterManagement = serviceCenterManagement;
        this.capInteractions = capInteractions;
    }

    public String processCommand(RMICommand command,
            Serializable... commandArguments) throws RemoteException,
            ApplicationException {
        String returnMessage = "success";
        ByteArrayInputStream bis;
        ACLMessage msg = new ACLMessage(ACLMessage.INFORM);

        GlobalLoopLogger.getLogger().debug(
                "Received command " + command.toString());
        switch (command) {



            case VM_ADDED_HOOK_ACTIVATED:
                if (GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                    for (Serializable arg : commandArguments) {
                        if (!(arg instanceof ActivityDto)) {
                            returnMessage += "VM_ADDED_HOOK_ACTIVATED expects service information";
                            throw new ApplicationException(
                                    "VM_ADDED_HOOK_ACTIVATED expects service information");
                        }

                        ActivityDto activityDto = (ActivityDto) arg;
                        try {
                            VirtualTaskInfo virtualTaskInfo = serviceCenterManagement.getVMInfo(activityDto.getCloudID());

                            returnMessage = processServiceDeploymentNotification(VirtualTaskInfoToActivityDtoMapper.map(virtualTaskInfo));
                        } catch (ServiceCenterAccessException e) {
                            throw new ApplicationException(e.toString(), e.getCause());
                        }

                    }
                }
                break;

            case VM_REMOVED_HOOK_ACTIVATED:
                if (GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                    for (Serializable arg : commandArguments) {
                        if (!(arg instanceof ActivityDto)) {
                            returnMessage += "VM_REMOVED_HOOK_ACTIVATED expects service information";
                            throw new ApplicationException(
                                    "VM_REMOVED_HOOK_ACTIVATED expects service information");
                        }
                        ActivityDto activityDto = (ActivityDto) arg;
                        try {
                            VirtualTaskInfo virtualTaskInfo = serviceCenterManagement.getVMInfo(activityDto.getCloudID());
                            returnMessage = processServiceTerminationNotification(VirtualTaskInfoToActivityDtoMapper.map(virtualTaskInfo));
                        } catch (ServiceCenterAccessException e) {
                            throw new ApplicationException(e.toString(), e.getCause());
                        }


                    }
                }
                break;

            /*
             * Includes a hack. Because the GCL also accesses CAP to write in the EPKB the deployed service,
             * after retrieving the information the GCL deltes from CAP the Service. 
             * Can cause problems later: ex. EPKB id's missmanagement
             */
            case JMS_SERVICE_DEPLOYMENT_NOTIFICATION:
                if (commandArguments == null || commandArguments.length == 0) {
                    returnMessage += "No argument specified to command";
                    throw new ApplicationException(
                            "No argument specified to command");
                }
                for (Serializable arg : commandArguments) {
                    if (!(arg instanceof ActivityDto)) {
                        returnMessage += "Service deployment notification expects service information";
                        throw new ApplicationException(
                                "Service deployment notification expects service information");
                    }

                    ActivityDto toBeDeployedInfo = (ActivityDto) arg;
                    if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                        ActivityDto existingService = capInteractions.getDeployedService(toBeDeployedInfo);
                        capInteractions.removeDeployedService(existingService);
                        if ((existingService == null || existingService.getEpkbID() == null)) {
                            GlobalLoopLogger.getLogger().debug("Service with epkbID="
                                    + toBeDeployedInfo.getEpkbID()
                                    + " requested for deployment was not found in the EPKB");
                            returnMessage = "Service not found";
                            return returnMessage;
                        }

                        existingService.setRequestedCores(toBeDeployedInfo.getRequestedCores());
                        existingService.setIp(toBeDeployedInfo.getIp());
                        existingService.setTaskName(toBeDeployedInfo.getTaskName());
                        existingService.setCpuWeight(0.3f);
                        existingService.setMemWeight(0.3f);
                        existingService.setHddWeight(0.3f);

                        returnMessage = processServiceDeploymentNotification(existingService);
                    }
                }

                break;

            case JMS_SERVICE_UNDEPLOYMENT_NOTIFICATION:
                if (commandArguments == null || commandArguments.length == 0) {
                    returnMessage += "No argument specified to command";
                    throw new ApplicationException(
                            "No argument specified to command");
                }
                for (Serializable arg : commandArguments) {
                    if (!(arg instanceof ActivityDto)) {
                        returnMessage += "Service termination notification expects service information";
                        throw new ApplicationException(
                                "Service termination notification expects service information");
                    }

                    ActivityDto toBeDeployedInfo = (ActivityDto) arg;
                    if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                        ActivityDto existingService = capInteractions.getDeployedService(toBeDeployedInfo);

                        if (existingService.getId() == null) {
                            GlobalLoopLogger.getLogger().info("Service with IP="
                                    + toBeDeployedInfo.getIp()
                                    + " requested to be terminated was not found in EPKB in service_instance");
                            returnMessage = "Service Instance not found";
                            return returnMessage;
                        }
                        existingService.setTaskName(toBeDeployedInfo.getTaskName());
                        returnMessage = processServiceTerminationNotification(existingService);
                    }

                }

                break;

            case SERVICE_DEPLOYMENT_NOTIFICATION:
                if (commandArguments == null || commandArguments.length == 0) {
                    returnMessage += "No argument specified to command";
                    throw new ApplicationException(
                            "No argument specified to command");
                }
                for (Serializable arg : commandArguments) {
                    if (!(arg instanceof ActivityDto)) {
                        returnMessage += "Service deployment notification expects service information";
                        throw new ApplicationException(
                                "Service deployment notification expects service information");
                    }

                    ActivityDto toBeDeployedInfo = (ActivityDto) arg;
                    if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                        ActivityDto existingService = capInteractions.getService(toBeDeployedInfo);

                        if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()
                                && (existingService == null || existingService.getEpkbID() == null)) {
                            GlobalLoopLogger.getLogger().debug("Service with epkbID="
                                    + toBeDeployedInfo.getEpkbID()
                                    + " requested for deployment was not found in the EPKB");
                            returnMessage = "Service not found";
                            return returnMessage;
                        }

                        existingService.setRequestedCores(toBeDeployedInfo.getRequestedCores());
                        existingService.setIp(toBeDeployedInfo.getIp());
//                  existingService.setLife(-1);
                        existingService.setTaskName(toBeDeployedInfo.getTaskName());
////                            + "_index_"
////                            + UIDGenerator.generateID(ActivityDto.class));
//                    existingService.setServiceName(toBeDeployedInfo.getServiceName());
////                    existingService.setId(toBeDeployedInfo.getId());
////                            + "_index_"
////                            + UIDGenerator.generateID(ApplicationActivity.class)
                        //existingService.setVmConfiguration(toBeDeployedInfo.getVmConfiguration());

                        // TODO: de citit de undeva dinamic CPU, MEM weight. Acuma sunt
                        // setate la 0.3 toate
                        existingService.setCpuWeight(0.3f);
                        existingService.setMemWeight(0.3f);
                        existingService.setHddWeight(0.3f);

                        returnMessage = processServiceDeploymentNotification(existingService);
                    } else {
                        toBeDeployedInfo.setCpuWeight(0.3f);
                        toBeDeployedInfo.setMemWeight(0.3f);
                        toBeDeployedInfo.setHddWeight(0.3f);

                        returnMessage = processServiceDeploymentNotification(toBeDeployedInfo);
                    }
                }
                break;
            case SERVICE_COLLECTION_DEPLOYMENT_NOTIFICATION:
                if (commandArguments == null || commandArguments.length == 0) {
                    returnMessage += "No argument specified to command";
                    throw new ApplicationException(
                            "No argument specified to command");
                } {
                if (!(commandArguments instanceof Serializable[])) {
                    returnMessage += "Service collection deployment notification expects service information";
                    throw new ApplicationException(
                            "Service collection deployment notification expects service information");
                }
                List<ActivityDto> toBeDeployed = new ArrayList<ActivityDto>();
                for (Serializable serializable : (Serializable[]) commandArguments) {
                    if (!(serializable instanceof ActivityDto)) {
                        returnMessage += "Service collection deployment notification expects service information";
                        GlobalLoopLogger.getLogger().error(returnMessage);
                        break;
                    }

                    ActivityDto toBeDeployedInfo = (ActivityDto) serializable;

                    GlobalLoopLogger.getLogger().info("Received command for deployment of :" + toBeDeployedInfo.getTaskName() + " " + toBeDeployedInfo.getIp());
                    if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {

                        ActivityDto existingService = capInteractions.getService(toBeDeployedInfo);

                        if ((existingService == null || existingService.getEpkbID() == null)) {
                            GlobalLoopLogger.getLogger().debug("Service with epkbID="
                                    + toBeDeployedInfo.getEpkbID()
                                    + " requested for deployment was not found in the EPKB");
                            returnMessage = "Service not found";
                            return returnMessage;
                        }


                        existingService.setRequestedCores(toBeDeployedInfo.getRequestedCores());
                        existingService.setIp(toBeDeployedInfo.getIp());
//                    existingService.setLife(-1);
                        existingService.setTaskName(toBeDeployedInfo.getTaskName());
////                            + "_index_"
////                            + UIDGenerator.generateID(ActivityDto.class));
//                    existingService.setServiceName(toBeDeployedInfo
//                            .getServiceName());
////                            + "_index_"
////                            + UIDGenerator
////                            .generateID(ApplicationActivity.class));
////                    existingService.setId(toBeDeployedInfo.getId());

                        // TODO: de citit de undeva dinamic CPU, MEM weight. Acuma
                        // sunt setate la 0.3 toate
                        existingService.setCpuWeight(0.3f);
                        existingService.setMemWeight(0.3f);
                        existingService.setHddWeight(0.3f);

                        toBeDeployed.add(existingService);
                    } else {
                        toBeDeployedInfo.setCpuWeight(0.3f);
                        toBeDeployedInfo.setMemWeight(0.3f);
                        toBeDeployedInfo.setHddWeight(0.3f);
                        toBeDeployed.add(toBeDeployedInfo);
                    }

                }
                returnMessage = processServiceCollectionDeploymentNotification(toBeDeployed);
            }


            break;
            case SERVICE_COLLECTION_UNDEPLOYMENT_NOTIFICATION:
                if (commandArguments == null || commandArguments.length == 0) {
                    returnMessage += "No argument specified to command";
                    throw new ApplicationException(
                            "No argument specified to command");
                }
                List<ActivityDto> toBeDeployed = new ArrayList<ActivityDto>();
                for (Serializable arg : commandArguments) {


                    if (!(arg instanceof ActivityDto)) {
                        returnMessage += "Service collection deployment notification expects service information";
                        throw new ApplicationException(
                                "Service collection deployment notification expects service information");
                    }

                    ActivityDto toBeDeployedInfo = (ActivityDto) arg;
                    GlobalLoopLogger.getLogger().debug("Received command for undeployment of: " + toBeDeployedInfo.getTaskName() + "_" + toBeDeployedInfo.getIp());

                    if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {

                        ActivityDto existingService = capInteractions.getDeployedService(toBeDeployedInfo);

                        if (existingService == null || existingService.getId() == null) {
                            GlobalLoopLogger.getLogger().info("Service with IP="
                                    + toBeDeployedInfo.getIp()
                                    + " requested to be terminated was not found in EPKB in service_instance");
                            returnMessage = "Service Instance not found";
                            continue;
//
// return returnMessage;
                        }
                        existingService.setTaskName(toBeDeployedInfo.getTaskName());
                        // existingService.setVmConfiguration(toBeDeployedInfo.getVmConfiguration());
                        toBeDeployed.add(existingService);
                    } else {
                        toBeDeployed.add(toBeDeployedInfo);
                    }

                }
                returnMessage = processServiceCollectionTerminationNotification(toBeDeployed);


                break;

            case SERVICE_UNDEPLOYMENT_NOTIFICATION:
                if (commandArguments == null || commandArguments.length == 0) {
                    returnMessage += "No argument specified to command";
                    throw new ApplicationException(
                            "No argument specified to command");
                }
                for (Serializable arg : commandArguments) {
                    if (!(arg instanceof ActivityDto)) {
                        returnMessage += "Service termination notification expects service information";
                        throw new ApplicationException(
                                "Service termination notification expects service information");
                    }

                    ActivityDto toBeDeployedInfo = (ActivityDto) arg;
                    if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                        ActivityDto existingService = capInteractions.getDeployedService(toBeDeployedInfo);

                        if (existingService.getId() == null) {
                            GlobalLoopLogger.getLogger().info("Service with IP="
                                    + toBeDeployedInfo.getIp()
                                    + " requested to be terminated was not found in EPKB in service_instance");
                            returnMessage = "Service Instance not found";
                            return returnMessage;
                        }
                        existingService.setTaskName(toBeDeployedInfo.getTaskName());
                        returnMessage = processServiceTerminationNotification(existingService);
                    } else {
                        returnMessage = processServiceTerminationNotification(toBeDeployedInfo);
                    }

                }

                break;

            case EXIT:
                InterAgentMessageBuilder.sendMessage(agent, MainAgent.class,
                        AgentMessageType.TERMINATE_EXECUTION);
                break;

            case ADD_APPLICATION:
                if (commandArguments == null || commandArguments.length == 0) {
                    returnMessage += "No argument specified to command";
                    throw new ApplicationException(
                            "No argument specified to command");
                }
                for (Serializable arg : commandArguments) {
                    if (!(arg instanceof byte[])) {
                        returnMessage += "Add application expects file";
                        GlobalLoopLogger.getLogger().debug(returnMessage);
                        throw new ApplicationException(
                                "application host expects file");
                    }
                    byte[] fileContent = (byte[]) arg;
                    bis = new ByteArrayInputStream(fileContent);

                    ApplicationDescription applicationDescription;
                    applicationDescription = JAXB.unmarshal(bis,
                            ApplicationDescription.class);
                    returnMessage = processAddApplicationCommand(new ApplicationDescription[]{applicationDescription});
                }
                break;
            case ADD_HOSTS:
                if (commandArguments == null || commandArguments.length == 0) {
                    returnMessage += "No argument specified to command";
                    throw new ApplicationException(
                            "No argument specified to command");
                }
                for (Serializable arg : commandArguments) {
                    if (!(arg instanceof byte[])) {
                        returnMessage += "Add host expects byte array";
                        throw new ApplicationException(
                                "Add host expects byte array");
                    }
                    byte[] fileContent = (byte[]) arg;
                    bis = new ByteArrayInputStream(fileContent);
                    ServerList serverList;
                    try {
                        serverList = JAXB.unmarshal(bis, ServerList.class);
                        returnMessage += processAddServerCommand(serverList.getServers());

                    } catch (Exception e) {
                        GlobalLoopLogger.getLogger().debug(e.toString());
                        returnMessage = "Specified file does not contain valid server description";
                        throw new ApplicationException(returnMessage);
                    }
                }
                break;

            case REMOVE_ACTIVITY:
                if (commandArguments == null || commandArguments.length == 0) {
                    returnMessage += "No argument specified to command";
                    throw new ApplicationException(
                            "No argument specified to command");
                }
                List<String> removeActivityNames = new ArrayList<String>();

                int namesCount = commandArguments.length;
                String[] names = new String[namesCount];
                for (int i = 0; i < namesCount; i++) {
                    names[i] = (String) commandArguments[i];
                }
                InterAgentMessageBuilder.sendMessage(agent,
                        ContextMonitoringAgent.class,
                        AgentMessageType.ACTIVITIES_REMOVED, names);

                break;
            case REMOVE_SERVICE:
                String removeServiceApplicationNames = (String) commandArguments[0];
                String removeServiceActivitiesNames = (String) commandArguments[1];
                String removeServiceServicesNames = (String) commandArguments[2];
                List<String> toBeRemoved = new ArrayList<String>();
                toBeRemoved.add(removeServiceApplicationNames + "_"
                        + removeServiceActivitiesNames + "_"
                        + removeServiceServicesNames);
                InterAgentMessageBuilder.sendMessage(agent,
                        ContextMonitoringAgent.class,
                        AgentMessageType.ACTIVITIES_REMOVED, toBeRemoved);
                break;
            case REMOVE_HOST:
                String removeHostName = (String) commandArguments[0];
                List<String> serversToBeRemoved = new ArrayList<String>();
                serversToBeRemoved.add(removeHostName);

                InterAgentMessageBuilder.sendMessage(agent,
                        ContextMonitoringAgent.class,
                        AgentMessageType.SERVERS_REMOVED, serversToBeRemoved);

                break;
           
            case READ_MONITORED_DATA:
                InterAgentMessageBuilder.sendMessage(agent, ContextMonitoringAgent.class, AgentMessageType.UPDATE_CONTEXT_INFORMATION);
                break;
                
            default:
                returnMessage = "Unrecognized command. Use:";
                returnMessage += "------------------------------";

                for (String commandName : RMICommand.getValues()) {
                    returnMessage += commandName;
                }
                returnMessage += "------------------------------";
                GlobalLoopLogger.getLogger().debug(returnMessage);
                break;
        }

        return returnMessage;
    }

    private String processAddServerCommand(Serializable[] args) {
        List<ExtendedServerDto> serverDtos = new ArrayList<ExtendedServerDto>();
        String message = "";
        for (Serializable arg : args) {
            ServerDescription serverDescription = (ServerDescription) arg;
            ServerPowerUsageDescription serverPowerUsageDescription = serverDescription.getServerPowerUsageDescription();
            // serverDescription.setHostname(serverDescription.getHostname() +
            // serverIndex++);
            serverDescription.setHostname(serverDescription.getHostname());
            boolean exists = false;

            String serverName = serverDescription.getHostname();
            ExtendedServerDto dto = new ExtendedServerDto();

            dto.setServerName(serverName);
            dto.setTmName(serverDescription.getTm());
            dto.setVmmName(serverDescription.getVmm());
            dto.setImName(serverDescription.getIm());
            dto.setIpAddress(serverDescription.getIp());
            dto.setMacAddress(serverDescription.getMac());
            dto.setMaximumCPU(serverDescription.getMaximumCpu());
            dto.setOptimumCPU(serverDescription.getOptimumCpu());
            dto.setMaximumMemory(serverDescription.getMaximumMem());
            dto.setOptimumMemory(serverDescription.getOptimumMem());
            dto.setCoreNo(serverDescription.getCoreCount());
            dto.setPowerConsumptionWhileIdle(serverPowerUsageDescription.getWhileIdle());
            dto.setPowerConsumptionWhileLoadedTo40(serverPowerUsageDescription.getWhileLoadedTo40());
            dto.setPowerConsumptionWhileLoadedTo100(serverPowerUsageDescription.getWhileLoadedTo100());

            List<StorageDto> storageDtos = new ArrayList<StorageDto>();
            for (ServerDiskDescription diskDescription : serverDescription.getServerDisks()) {
                StorageDto storageDto = new StorageDto();
                storageDto.setMaximumStorage(diskDescription.getMaximumStorage());
                storageDto.setOptimumStorage(diskDescription.getOptimumStorage());
                storageDto.setName(diskDescription.getTarget());
            }
            dto.setStorageDtos(storageDtos);
            serverDtos.add(dto);
        }

        message += sendServersAddedMessage(serverDtos);

        return message;
    }

    private String processAddApplicationCommand(Serializable[] args) {
        String message = "";
        List<ApplicationDto> applications = new ArrayList<ApplicationDto>();
        for (Serializable arg : args) {
            ApplicationDescription applicationDescription = (ApplicationDescription) arg;
            // applicationDescription.setName(applicationDescription.getName() +
            // taskIndex++);
            applicationDescription.setName(applicationDescription.getName());
            boolean exists = false;

            ApplicationDto applicationDto = new ApplicationDto(
                    applicationDescription.getName());
            for (ActivityDescription activityDescription : applicationDescription.getActivities()) {
                for (ActivityService activityService : activityDescription.getServices()) {
                    ActivityDto task = new ActivityDto();
                    task.setTaskName(activityDescription.getName());
                    task.setServiceName(activityService.getServiceName());
                    task.setRequestedCores(activityService.getRequestedCoresCount());
                    task.setCpuWeight(activityService.getCpuWeight());
                    task.setMemWeight(activityService.getMemWeight());
                    task.setHddWeight(activityService.getStorageWeight());
                    task.setRequestedCPUMin(activityService.getRequestedCPUMin());
                    task.setRequestedCPUMax(activityService.getRequestedCPUMax());
                    task.setRequestedMemoryMin(activityService.getRequestedMemoryMin());
                    task.setRequestedMemoryMax(activityService.getRequestedMemoryMax());
                    task.setRequestedStorageMin(activityService.getRequestedStorageMin());
                    task.setRequestedStorageMax(activityService.getRequestedStorageMax());
                    task.setVncPort(activityService.getVncPort());
                    task.setIp(activityService.getIp());
                    task.setLife(activityDescription.getDuration());
                    List<ActivityDiskDto> taskDiskDtos = new ArrayList<ActivityDiskDto>();
                    for (VirtualDiskDescription virtualDiskDescription : activityService.getDisks()) {
                        ActivityDiskDto taskDiskDto = new ActivityDiskDto();
                        taskDiskDto.setType(virtualDiskDescription.getType());
                        taskDiskDto.setFormat(virtualDiskDescription.getFormat());
                        taskDiskDto.setSize(virtualDiskDescription.getSize());
                        taskDiskDto.setPath(virtualDiskDescription.getSource());
                        taskDiskDtos.add(taskDiskDto);
                    }
                    task.setDiskDtoCollection(taskDiskDtos);
                    applicationDto.addActivity(task);
                }
                applications.add(applicationDto);
            }

        }

        message += sendTasksAddedMessage(applications);
        return message;
    }

    private void processAddScheduleCommand(Serializable[] args) {
        for (Serializable arg : args) {
            WorkloadSchedule schedule = (WorkloadSchedule) arg;
            for (final JaxbPair pair : schedule.getApplications()) {
                Timer timer = new Timer();
                TimerTask task = new TimerTask() {
                    //

                    public void run() {
                        processAddApplicationCommand(new Serializable[]{pair.getApplication()});
                        // GlobalLoopLogger.getLogger().info("===========================");
                        // GlobalLoopLogger.getLogger().info("===========================");
                        // GlobalLoopLogger.getLogger().info("===========================");
                        // GlobalLoopLogger.getLogger().info(new Date());
                        // GlobalLoopLogger.getLogger().info("===========================");
                        // GlobalLoopLogger.getLogger().info("===========================");
                        // GlobalLoopLogger.getLogger().info("===========================");
                        // System.err.println(new Date());
                    }
                };
                timer.schedule(task, pair.getStartDelay() * 1000);
            }
        }
    }

    private void processAddScheduleWithoutCheckForDuplicatesCommand(
            Serializable[] args) {
        for (Serializable arg : args) {
            WorkloadSchedule schedule = (WorkloadSchedule) arg;
            for (final JaxbPair pair : schedule.getApplications()) {
                Timer timer = new Timer();
                TimerTask task = new TimerTask() {

                    public void run() {
                        processAddApplicationWithoutCheckForDuplicatesCommand(new Serializable[]{pair.getApplication()});
                    }
                };
                timer.schedule(task, pair.getStartDelay());
            }
        }
    }

    private String processAddApplicationWithoutCheckForDuplicatesCommand(
            Serializable[] args) {

        List<ApplicationDto> applications = new ArrayList<ApplicationDto>();

        String message = "";
        for (Serializable arg : args) {
            ApplicationDescription applicationDescription = (ApplicationDescription) arg;

            for (ActivityDescription activityDescription : applicationDescription.getActivities()) {

                ApplicationDto applicationDto = new ApplicationDto(
                        activityDescription.getName());
                for (ActivityService activityService : activityDescription.getServices()) {
                    ActivityDto task = new ActivityDto();
                    task.setTaskName(applicationDescription.getName());// + "_"
                    // +
                    // activityDescription.getName()
                    // + "_"
                    // +
                    // activityService.getServiceName());
                    task.setRequestedCores(activityService.getRequestedCoresCount());
                    task.setCpuWeight(activityService.getCpuWeight());
                    task.setMemWeight(activityService.getMemWeight());
                    task.setHddWeight(activityService.getStorageWeight());
                    task.setRequestedCPUMin(activityService.getRequestedCPUMin());
                    task.setRequestedCPUMax(activityService.getRequestedCPUMax());
                    task.setRequestedMemoryMin(activityService.getRequestedMemoryMin());
                    task.setRequestedMemoryMax(activityService.getRequestedMemoryMax());
                    task.setRequestedStorageMin(activityService.getRequestedStorageMin());
                    task.setRequestedStorageMax(activityService.getRequestedStorageMax());
                    task.setVncPort(activityService.getVncPort());
                    task.setIp(activityService.getIp());
                    task.setLife(activityDescription.getDuration());
                    List<ActivityDiskDto> taskDiskDtos = new ArrayList<ActivityDiskDto>();
                    for (VirtualDiskDescription virtualDiskDescription : activityService.getDisks()) {
                        ActivityDiskDto taskDiskDto = new ActivityDiskDto();
                        taskDiskDto.setType(virtualDiskDescription.getType());
                        taskDiskDto.setFormat(virtualDiskDescription.getFormat());
                        taskDiskDto.setSize(virtualDiskDescription.getSize());
                        taskDiskDto.setPath(virtualDiskDescription.getSource());
                        taskDiskDtos.add(taskDiskDto);
                    }
                    task.setDiskDtoCollection(taskDiskDtos);
                    applicationDto.addActivity(task);
                }
                applications.add(applicationDto);
            }

        }

        message += sendTasksAddedMessage(applications);
        return message;
    }

    private boolean checkFileCommand(String params) {

        File file = new File(params);
        if (!file.exists()) {
            GlobalLoopLogger.getLogger().info("------------------------------");
            GlobalLoopLogger.getLogger().info("Inexistent file specified");
            GlobalLoopLogger.getLogger().info("------------------------------");
            return false;
        }
        return true;
    }

    private String sendTasksAddedMessage(List<ApplicationDto> dtos) {
        String message = "";
        InterAgentMessageBuilder.sendMessage(agent, ContextMonitoringAgent.class,
                AgentMessageType.UPDATE_EACM_CONCEPTS, AgentMessageType.ACTIVITIES_ADDED, dtos);
        message += "\nsuccess";

        return message;
    }

    private String sendServersAddedMessage(List<ExtendedServerDto> serverDtos) {
        String result = "\nsuccess";
        InterAgentMessageBuilder.sendMessage(agent, ContextMonitoringAgent.class, AgentMessageType.UPDATE_EACM_CONCEPTS,
                AgentMessageType.SERVERS_ADDED, serverDtos);
        return result;
    }

    private String processServiceCollectionDeploymentNotification(
            List<ActivityDto> dtos) {
        String message = "";
        if ((GeneralConfigurationManager.getPlanningEngine().equals("ffd") || GeneralConfigurationManager.getPlanningEngine().equals("rl"))) {
            List<ApplicationDto> applicationDtos = new ArrayList<ApplicationDto>(
                    dtos.size());

            for (ActivityDto dto : dtos) {
                ApplicationDto applicationDto = new ApplicationDto("Application_"
                        + taskIndex++);
                applicationDto.addActivity(dto);
                applicationDtos.add(applicationDto);
                GlobalLoopLogger.getLogger().debug(
                        "Service with epkbID=" + dto.getEpkbID()
                        + " sent for deployment");
            }

            message = sendTasksAddedMessage(applicationDtos);
        } else {
            //here it enters when the OpenNebula mm_sched must deploy the VM
            for (ActivityDto dto : dtos) {
                VirtualNetworkInfo network = new VirtualNetworkInfo();


                VirtualTaskInfo taskInfo = new VirtualTaskInfo(dto.getTaskName());

                if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {
                    //TODO: to be fixed. Sa nu citesc chiar asa direct template cu path sau ceva se tat acolo. k acuma in EPKNB i pathu

                    String vmTemplate = "";
                    BufferedReader reader = null;
                    try {
                        reader = new BufferedReader(new FileReader(dto.getVmConfiguration()));
                        String line = "";
                        while ((line = reader.readLine()) != null) {
                            vmTemplate += line + "\n";
                        }
                    } catch (Exception e) {
                        GlobalLoopLogger.getLogger().info(e.toString());
                        GlobalLoopLogger.error(e);
                    }

                    taskInfo.setVmTemplate(vmTemplate);
                    network.setIp(dto.getIp());
                    network.setVncPassword("123456");
                    network.setVncPort(dto.getVncPort());
                    taskInfo.setNetworkInfo(network);
                    for (ActivityDiskDto activityDiskDto : dto.getDiskDtoCollection()) {
                        VirtualDiskInfo virtualDiskInfo = new VirtualDiskInfo();
                        virtualDiskInfo.setFormat(activityDiskDto.getFormat());
                        virtualDiskInfo.setSize(activityDiskDto.getSize());
                        virtualDiskInfo.setSource(activityDiskDto.getPath());
                        virtualDiskInfo.setType(activityDiskDto.getType());
                        taskInfo.addDiskDto(virtualDiskInfo);
                    }
                }


                taskInfo.setRequestedCPU(dto.getRequestedCPUMax());
                taskInfo.setRequestedMemory(dto.getRequestedMemoryMax());
                try {
                    VirtualTaskInfo virtualTaskInfo = serviceCenterManagement.createVirtualMachine(taskInfo);
                    dto.setCloudID(virtualTaskInfo.getId());
                    capInteractions.addDeployedService(dto);

                } catch (ServiceCenterAccessException e) {
                    GlobalLoopLogger.getLogger().info(e.toString());
                    GlobalLoopLogger.error(e);
                }
            }

            if (GeneralConfigurationManager.usesWorkloadGenerator()) {
                InterAgentMessageBuilder.sendMessage(agent, ro.utcn.dsrl.workloadScheduler.WorkloadSchedulerAgent.class, AgentMessageType.ACTION_PLAN_EXECUTION_COMPLETED);
            }
        }

        return message;
    }

    private String processServiceDeploymentNotification(ActivityDto dto) {
        String message = "";
        if ((GeneralConfigurationManager.getPlanningEngine().equals("ffd") || GeneralConfigurationManager.getPlanningEngine().equals("rl"))) {
            ApplicationDto applicationDto = new ApplicationDto("Application_"
                    + taskIndex++);
            applicationDto.addActivity(dto);

            List<ApplicationDto> applicationDtos = new ArrayList<ApplicationDto>();
            applicationDtos.add(applicationDto);

            message = sendTasksAddedMessage(applicationDtos);
            GlobalLoopLogger.getLogger().debug(
                    "Service with epkbID=" + dto.getEpkbID()
                    + " sent for deployment");
        } else {

            VirtualNetworkInfo network = new VirtualNetworkInfo();


            VirtualTaskInfo taskInfo = new VirtualTaskInfo(dto.getTaskName());

            if (!GeneralConfigurationManager.isDeployedAsOpenNebulaScheduler()) {

                String vmTemplate = "";
                BufferedReader reader = null;
                try {
                    reader = new BufferedReader(new FileReader(dto.getVmConfiguration()));
                    String line = "";
                    while ((line = reader.readLine()) != null) {
                        vmTemplate += line + "\n";
                    }
                } catch (Exception e) {
                    GlobalLoopLogger.getLogger().info(e.toString());
                    GlobalLoopLogger.error(e);
                }

                taskInfo.setVmTemplate(vmTemplate);
                network.setIp(dto.getIp());
                network.setVncPassword("123456");
                network.setVncPort(dto.getVncPort());
                taskInfo.setNetworkInfo(network);
                for (ActivityDiskDto activityDiskDto : dto.getDiskDtoCollection()) {
                    VirtualDiskInfo virtualDiskInfo = new VirtualDiskInfo();
                    virtualDiskInfo.setFormat(activityDiskDto.getFormat());
                    virtualDiskInfo.setSize(activityDiskDto.getSize());
                    virtualDiskInfo.setSource(activityDiskDto.getPath());
                    virtualDiskInfo.setType(activityDiskDto.getType());
                    taskInfo.addDiskDto(virtualDiskInfo);
                }
            }


            taskInfo.setRequestedCPU(dto.getRequestedCPUMax());
            taskInfo.setRequestedMemory(dto.getRequestedMemoryMax());
            try {
                capInteractions.addDeployedService(dto);
                serviceCenterManagement.createVirtualMachine(taskInfo);
            } catch (ServiceCenterAccessException e) {
                GlobalLoopLogger.getLogger().info(e.toString());
                GlobalLoopLogger.error(e);
            }


            if (GeneralConfigurationManager.usesWorkloadGenerator()) {
                InterAgentMessageBuilder.sendMessage(agent, ro.utcn.dsrl.workloadScheduler.WorkloadSchedulerAgent.class, AgentMessageType.ACTION_PLAN_EXECUTION_COMPLETED);
            }
        }

        return message;
    }

    private String processServiceTerminationNotification(ActivityDto dto) {
        String message = "";
        if ((GeneralConfigurationManager.getPlanningEngine().equals("ffd") || GeneralConfigurationManager.getPlanningEngine().equals("rl"))) {
            List<ActivityDto> dtos = new ArrayList<ActivityDto>();
            dtos.add(dto);

            message += "\nsuccess";

            InterAgentMessageBuilder.sendMessage(agent, ContextMonitoringAgent.class, AgentMessageType.UPDATE_EACM_CONCEPTS,
                    AgentMessageType.ACTIVITIES_REMOVED, dtos);
        } else {
            VirtualTaskInfo virtualTaskInfo = new VirtualTaskInfo(dto.getTaskName());
            virtualTaskInfo.setId(Integer.parseInt(dto.getId()));
            try {
                capInteractions.removeDeployedService(dto);
                serviceCenterManagement.removeVirtualMachine(virtualTaskInfo);
            } catch (ServiceCenterAccessException e) {
                GlobalLoopLogger.getLogger().info(e.toString());
                GlobalLoopLogger.error(e);
            }
            if (GeneralConfigurationManager.usesWorkloadGenerator()) {
                InterAgentMessageBuilder.sendMessage(agent, ro.utcn.dsrl.workloadScheduler.WorkloadSchedulerAgent.class, AgentMessageType.ACTION_PLAN_EXECUTION_COMPLETED);
            }
        }
        return message;
    }

    private String processServiceCollectionTerminationNotification(
            List<ActivityDto> dtos) {

        String message = "";
        if ((GeneralConfigurationManager.getPlanningEngine().equals("ffd") || GeneralConfigurationManager.getPlanningEngine().equals("rl"))) {
            message += "\nsuccess";

            InterAgentMessageBuilder.sendMessage(agent, ContextMonitoringAgent.class, AgentMessageType.UPDATE_EACM_CONCEPTS,
                    AgentMessageType.ACTIVITIES_REMOVED, dtos);
        } else {
            for (ActivityDto activityDto : dtos) {
                VirtualTaskInfo virtualTaskInfo = new VirtualTaskInfo(activityDto.getTaskName());
                virtualTaskInfo.setId(activityDto.getCloudID());
                try {
                    ActivityDto deployedServiceInfo = capInteractions.getDeployedService(activityDto);
                    capInteractions.removeDeployedService(deployedServiceInfo);
                    virtualTaskInfo.setId(deployedServiceInfo.getCloudID());
                    serviceCenterManagement.removeVirtualMachine(virtualTaskInfo);
                } catch (ServiceCenterAccessException e) {
                    GlobalLoopLogger.getLogger().info(e.toString());
                    GlobalLoopLogger.error(e);
                }
            }
            if (GeneralConfigurationManager.usesWorkloadGenerator()) {
                InterAgentMessageBuilder.sendMessage(agent, ro.utcn.dsrl.workloadScheduler.WorkloadSchedulerAgent.class, AgentMessageType.ACTION_PLAN_EXECUTION_COMPLETED);
            }
        }
        return message;
    }
}