/**********************************************************************************************************
 * Copyright 2012, Distributed Systems Research Laboratory, Technical University of Cluj-Napoca, Romania 
 * http://dsrl.coned.utcluj.ro/
 *  
 * Licensed under the EUPL V.1.1 
 * 
 * European Union Public Licence V. 1.1  
 * You may obtain a copy of this license at
 * 
 * http://joinup.ec.europa.eu/software/page/eupl/licence-eupl
 *  
 **********************************************************************************************************/
package ro.utcn.dsrl.utils.similarContextIdentification;


import ro.utcn.dsrl.utils.commonDtos.ContextSituationDto;
import ro.utcn.dsrl.utils.commonDtos.ActivityDto;
import ro.utcn.dsrl.utils.commonDtos.ExtendedServerDto;
import ro.utcn.dsrl.utils.misc.Pair;

import java.util.*;

/**
 * <br/>
 * <br/>
 * <p><strong>GAMES Project - Global Control Loop</strong></p>
 * <p><strong>@Author: Technical University of Cluj-Napoca</strong></p>
 * <br/>
 * <a href=http://dsrl.coned.utcluj.ro/ > Distributed Systems research Laboratory </a>
 * <br/>Coordinator: Prof. dr. ing. Ioan Salomie E-mail: Ioan.Salomie@cs.utcluj.ro
 * <br/>
 * <br/>Developers:
 * <br/>Daniel Moldovan(Daniel.Moldovan@cs.utcluj.ro)
 * <br/>Georgiana Copil(Georgiana.Copil@cs.utcluj.ro)
 */


public class SimilarContextSituationIdentification {
    private static ClusteringAlgorithm clusteringAlgorithmForTasks;
    private static ClusteringAlgorithm clusteringAlgorithmForServers;
    private static ClusteringAlgorithm clusteringAlgorithmForContexts;
    public static final int COMPARISON_ACCURACY_FOR_TASKS = 20;
    public static final int COMPARISON_ACCURACY_FOR_SERVERS = 10;
    public static final int COMPARISON_ACCURACY_FOR_CONTEXTS = 10;
    private static SimilarContextSituationIdentification similarContextSituationIdentification;

    public SimilarContextSituationIdentification() {
        ClusteringAlgorithmFactory factory = new ClusteringAlgorithmFactory();
        clusteringAlgorithmForTasks = factory.getKMeansAlgorithm(COMPARISON_ACCURACY_FOR_TASKS);
        clusteringAlgorithmForServers = factory.getKMeansAlgorithm(COMPARISON_ACCURACY_FOR_SERVERS);
        clusteringAlgorithmForContexts = factory.getKMeansAlgorithm(COMPARISON_ACCURACY_FOR_CONTEXTS);
    }

    public List<Cluster> getEnergyClusters() {
        return clusteringAlgorithmForContexts.getAllClusters();
    }

    /**
     * @param contextSituationDto pair <ActionFromCurrentContextName,ActionFromSavedContextName> =
     *                            match between action that can be executed in current context and action found in saved context
     * @return
     */
    public Map<String, String> matchTasks(ContextSituationDto contextSituationDto, List<ActivityDto> tasks) {
        HashMap<String, String> associatedTask = new HashMap<String, String>();

        for (ActivityDto dto : contextSituationDto.getTasks()) {

            for (ActivityDto dto1 : tasks) {
                if (!associatedTask.containsKey(dto.getTaskName())) {
                    if (clusteringAlgorithmForTasks.getBelongingCluster(dto) ==
                            clusteringAlgorithmForTasks.getBelongingCluster(dto1)
                            && !associatedTask.containsValue(dto1.getTaskName())) {
                        associatedTask.put(dto.getTaskName(), dto1.getTaskName());
                    }
                }
            }
        }
        return associatedTask;
    }

    /**
     * @param contextSituationDto pair <ServerFromCurrentContext,ServerFromSavedContext> =  match between
     *                            curent contetx servers and saved state servers
     * @return
     */
    public Map<String, String> matchServers(ContextSituationDto contextSituationDto, List<ExtendedServerDto> servers) {
        HashMap<String, String> associatedServer = new HashMap<String, String>();

        for (ExtendedServerDto dto : contextSituationDto.getServers()) {
            for (ExtendedServerDto dto1 : servers) {
                if (!associatedServer.containsKey(dto.getServerName())) {
                    if ((clusteringAlgorithmForServers.getBelongingCluster(dto) == clusteringAlgorithmForServers.getBelongingCluster(dto1)) && !associatedServer.containsValue(dto1.getServerName())) {
                        //GlobalLoopLogger.getLogger().info("Match: " + dto.getServerName() + " and " + dto1.getServerName());
                        associatedServer.put(dto1.getServerName(), dto.getServerName());
                    }
                }
            }
        }
        return associatedServer;
    }


    public void addTasks(List<ActivityDto> tasks) {
        if (clusteringAlgorithmForTasks.getNrOfClusters() == 0)
            clusteringAlgorithmForTasks.initializeClusters(tasks, ActivityDto.class);
        else clusteringAlgorithmForTasks.addObjectsToKnowledgeBase(tasks);

    }

    public void addServers(List<ExtendedServerDto> servers) {
        if (clusteringAlgorithmForServers.getNrOfClusters() == 0) {
            clusteringAlgorithmForServers.initializeClusters(servers, ExtendedServerDto.class);
        } else {
            clusteringAlgorithmForServers.addObjectsToKnowledgeBase(servers);
        }

    }

    public void initializeEnergyClusters(List<Cluster> clusters, Collection<ContextSituationDto> situations) {
        if (clusteringAlgorithmForContexts.getNrOfClusters() == 0) {
            clusteringAlgorithmForContexts.setInitialClusters(clusters);
            clusteringAlgorithmForContexts.initializeClusters(situations, ContextSituationDto.class);
        } else {
            clusteringAlgorithmForContexts.addObjectsToKnowledgeBase(situations);
        }
        clusteringAlgorithmForContexts.refreshClusters();
    }

    public void addResourcesFromSituations(Collection<ContextSituationDto> situations) {
        if (situations.size() > 0)
            for (ContextSituationDto situation : situations) {
                clusteringAlgorithmForServers.addObjectsToKnowledgeBase(situation.getServers());
                clusteringAlgorithmForTasks.addObjectsToKnowledgeBase(situation.getTasks());
                clusteringAlgorithmForServers.refreshClusters();
                clusteringAlgorithmForTasks.refreshClusters();
            }
    }

    public boolean checkIfContextsSimilar(ContextSituationDto currentContext, ContextSituationDto existingContext) {

        int tasksC1[] = new int[clusteringAlgorithmForTasks.getNrOfClusters()];
        int tasksC2[] = new int[clusteringAlgorithmForTasks.getNrOfClusters()];
        int nr = 0;
        for (ActivityDto dto : currentContext.getTasks()) {
            nr = clusteringAlgorithmForTasks.getBelongingCluster(dto);
            if (nr > -1)
                tasksC1[nr]++;
        }
        for (ActivityDto dto : existingContext.getTasks()) {
            nr = clusteringAlgorithmForTasks.getBelongingCluster(dto);
            if (nr > -1)
                tasksC2[nr]++;
        }
        boolean different = false;
        for (int i = 0; i < clusteringAlgorithmForTasks.getNrOfClusters(); i++) {
            if (tasksC1[i] != tasksC2[i]) {
                different = true;
            }
        }
        if (!different) {
            int serversC1[] = new int[clusteringAlgorithmForServers.getNrOfClusters()];
            int serversC2[] = new int[clusteringAlgorithmForServers.getNrOfClusters()];
            nr = 0;
            for (ExtendedServerDto dto : currentContext.getServers()) {
                nr = clusteringAlgorithmForServers.getBelongingCluster(dto);
                if (nr > -1) {
                    serversC1[nr]++;
                }
            }
            for (ExtendedServerDto dto : existingContext.getServers()) {
                nr = clusteringAlgorithmForServers.getBelongingCluster(dto);
                if (nr > -1) {
                    serversC2[nr]++;
                }
            }
            for (int i = 0; i < clusteringAlgorithmForTasks.getNrOfClusters(); i++) {
                if (serversC1[i] != serversC1[i]) {
                    different = true;
                }
            }
        }
        return !different;

    }

    public void refreshEnergyClusters() {

        clusteringAlgorithmForContexts.refreshClusters();

    }

    public void addSituationsToEnergyClusters(Collection<ContextSituationDto> situations) {
        if (situations.size() == 0) return;
        clusteringAlgorithmForContexts.addObjectsToKnowledgeBase(situations);
        clusteringAlgorithmForContexts.refreshClusters();
    }

    public void addSituationToEnergyClusters(ContextSituationDto situation) {
        if (situation == null) return;
        clusteringAlgorithmForContexts.addObjectToKnowledgeBase(situation);
        clusteringAlgorithmForContexts.refreshClusters();
    }

    public void createEnergyClusters(Collection<ContextSituationDto> situations) {

        if (clusteringAlgorithmForContexts.getNrOfClusters() == 0) {
            clusteringAlgorithmForContexts.initializeClusters(situations, ContextSituationDto.class);
        } else {
            clusteringAlgorithmForContexts.addObjectsToKnowledgeBase(situations);
        }
        clusteringAlgorithmForContexts.refreshClusters();
    }

    public List<Pair<ContextSituationDto, Double>> getOrderedListOfContextFromTheNearestEnergyCluster(ContextSituationDto currentContext, double currentEnergy) {
        List<Pair<ContextSituationDto, Double>> pairs = new ArrayList<Pair<ContextSituationDto, Double>>();
        List<ActivityDto> tasks = currentContext.getTasks();
        List<ExtendedServerDto> servers = currentContext.getServers();

        Cluster currentCluster = clusteringAlgorithmForContexts.getNearestCluster(currentEnergy);

        if (currentCluster == null) return pairs;
        for (Object object : currentCluster.getAllElements()) {
            if (object instanceof ContextSituationDto) {
                ContextSituationDto contextSituationDto = (ContextSituationDto) object;
                if (tasks.size() != contextSituationDto.getTasks().size() || servers.size() != contextSituationDto.getServers().size())
                    pairs.add(new Pair<ContextSituationDto, Double>(contextSituationDto, ContextSituationDto.MAXIMUM_DISTANCE));
                else {
                    clusteringAlgorithmForTasks.addObjectsToKnowledgeBase(contextSituationDto.getTasks());
                    clusteringAlgorithmForServers.addObjectsToKnowledgeBase(contextSituationDto.getServers());
                    clusteringAlgorithmForTasks.refreshClusters();
                    clusteringAlgorithmForServers.refreshClusters();
                    int tasksC1[] = new int[clusteringAlgorithmForTasks.getNrOfClusters()];
                    int tasksC2[] = new int[clusteringAlgorithmForTasks.getNrOfClusters()];
                    int nr = 0;
                    for (ActivityDto dto : tasks) {
                        nr = clusteringAlgorithmForTasks.getBelongingCluster(dto);
                        if (nr > -1)
                            tasksC1[nr]++;
                    }
                    for (ActivityDto dto : contextSituationDto.getTasks()) {
                        nr = clusteringAlgorithmForTasks.getBelongingCluster(dto);
                        if (nr > -1)
                            tasksC2[nr]++;
                    }
                    boolean different = false;
                    for (int i = 0; i < clusteringAlgorithmForTasks.getNrOfClusters(); i++) {
                        if (tasksC1[i] != tasksC2[i]) {
                            different = true;
                        }
                    }
                    if (!different) {
                        int serversC1[] = new int[clusteringAlgorithmForServers.getNrOfClusters()];
                        int serversC2[] = new int[clusteringAlgorithmForServers.getNrOfClusters()];
                        nr = 0;
                        for (ExtendedServerDto dto : servers) {
                            nr = clusteringAlgorithmForServers.getBelongingCluster(dto);
                            if (nr > -1) {
                                serversC1[nr]++;
                            }
                        }
                        for (ExtendedServerDto dto : contextSituationDto.getServers()) {
                            nr = clusteringAlgorithmForServers.getBelongingCluster(dto);
                            if (nr > -1) {
                                serversC2[nr]++;
                            }
                        }
                        for (int i = 0; i < clusteringAlgorithmForTasks.getNrOfClusters(); i++) {
                            //Daca contextul actual are mai multe servere decat contextul salvat intr-un cluster nu e nici o problema
                            if (serversC1[i] != serversC2[i] && serversC1[i] > serversC2[i]) {
                                different = true;
                            }
                        }

                    }
                    if (different == true)
                        pairs.add(new Pair<ContextSituationDto, Double>(contextSituationDto, ContextSituationDto.MAXIMUM_DISTANCE));
                    else
                        pairs.add(new Pair<ContextSituationDto, Double>(contextSituationDto, currentContext.distance(contextSituationDto)));
                }
            }
        }
        return order(pairs);
    }

    /**
     * TODO: de modificat sa returnam numai situatiile de context din acelasi cluster energetic in ordinea crescatoare a distantei
     */
    public List<Pair<ContextSituationDto, Double>> getOrderedListOfContexts(List<ContextSituationDto> knowledgeBase, ContextSituationDto currentContext) {
        List<Pair<ContextSituationDto, Double>> pairs = new ArrayList<Pair<ContextSituationDto, Double>>();
        List<ActivityDto> tasks = currentContext.getTasks();
        List<ExtendedServerDto> servers = currentContext.getServers();
        for (ContextSituationDto contextSituationDto : knowledgeBase) {
            if (tasks.size() != contextSituationDto.getTasks().size() || servers.size() != contextSituationDto.getServers().size())
                pairs.add(new Pair<ContextSituationDto, Double>(contextSituationDto, ContextSituationDto.MAXIMUM_DISTANCE));
            else {
                clusteringAlgorithmForTasks.addObjectsToKnowledgeBase(contextSituationDto.getTasks());
                clusteringAlgorithmForServers.addObjectsToKnowledgeBase(contextSituationDto.getServers());
                clusteringAlgorithmForTasks.refreshClusters();
                clusteringAlgorithmForServers.refreshClusters();
                int tasksC1[] = new int[clusteringAlgorithmForTasks.getNrOfClusters()];
                int tasksC2[] = new int[clusteringAlgorithmForTasks.getNrOfClusters()];
                int nr = 0;
                for (ActivityDto dto : tasks) {
                    nr = clusteringAlgorithmForTasks.getBelongingCluster(dto);
                    if (nr > -1)
                        tasksC1[nr]++;
                }
                for (ActivityDto dto : contextSituationDto.getTasks()) {
                    nr = clusteringAlgorithmForTasks.getBelongingCluster(dto);
                    if (nr > -1)
                        tasksC2[nr]++;
                }
                boolean different = false;
                for (int i = 0; i < clusteringAlgorithmForTasks.getNrOfClusters(); i++) {
                    if (tasksC1[i] != tasksC2[i]) {
                        different = true;
                    }
                }
                if (!different) {
                    int serversC1[] = new int[clusteringAlgorithmForServers.getNrOfClusters()];
                    int serversC2[] = new int[clusteringAlgorithmForServers.getNrOfClusters()];
                    nr = 0;
                    for (ExtendedServerDto dto : servers) {
                        nr = clusteringAlgorithmForServers.getBelongingCluster(dto);
                        if (nr > -1) {
                            serversC1[nr]++;
                        }
                    }
                    for (ExtendedServerDto dto : contextSituationDto.getServers()) {
                        nr = clusteringAlgorithmForServers.getBelongingCluster(dto);
                        if (nr > -1) {
                            serversC2[nr]++;
                        }
                    }
                    for (int i = 0; i < clusteringAlgorithmForTasks.getNrOfClusters(); i++) {
                        //Daca contextul actual are mai multe servere decat contextul salvat intr-un cluster nu e nici o problema
                        if (serversC1[i] != serversC2[i] && serversC1[i] > serversC2[i]) {
                            different = true;
                        }
                    }

                }
                if (different == true)
                    pairs.add(new Pair<ContextSituationDto, Double>(contextSituationDto, ContextSituationDto.MAXIMUM_DISTANCE));
                else
                    pairs.add(new Pair<ContextSituationDto, Double>(contextSituationDto, currentContext.distance(contextSituationDto)));
            }
        }
        return order(pairs);
    }

    public static List<Pair<ContextSituationDto, Double>> order(List<Pair<ContextSituationDto, Double>> pairs) {
        //TODO: better ordering in here
        for (int i = 0; i < pairs.size(); i++)
            for (int j = i + 1; j < pairs.size(); j++) {
                if (pairs.get(i).getSecond() > pairs.get(j).getSecond()) {
                    Pair<ContextSituationDto, Double> pair = pairs.get(i);
                    pairs.set(i, pairs.get(j));
                    pairs.set(j, pair);
                }
            }
        return pairs;
    }

    public static SimilarContextSituationIdentification getIdentification() {
        if (similarContextSituationIdentification == null) {
            similarContextSituationIdentification = new SimilarContextSituationIdentification();
        }
        return similarContextSituationIdentification;
    }

    public void refresh() {
        clusteringAlgorithmForTasks.refreshClusters();
        clusteringAlgorithmForServers.refreshClusters();
        clusteringAlgorithmForContexts.refreshClusters();
    }
}
